"""Helper for updating a projectum's log_url
Log_urls are the google doc links to a projectum's Projectum Agenda Log
"""

from regolith.fsclient import _id_key
from regolith.helpers.basehelper import DbHelperBase
from regolith.tools import all_docs_from_collection, fragment_retrieval

TARGET_COLL = "projecta"


def subparser(subpi):
    subpi.add_argument(
        "projectum_id",
        help="the ID or fragment of the ID of the " "Projectum to be updated, e.g., " "vl_m.",
        default=None,
    )
    subpi.add_argument("log_url", help="Google Doc url link to project's Projectum Agenda Log")
    subpi.add_argument(
        "-i",
        "--index",
        help="The index of the id you would like " "to update, from the returned list " "of projectum ids.",
    )
    # Do not delete --database arg
    subpi.add_argument(
        "-d",
        "--database",
        help="The database that will be updated.  Defaults to " "first database in the regolithrc.json file.",
    )
    return subpi


class LogUrlUpdaterHelper(DbHelperBase):
    """
    Update a projectum's Log_url, will add a new Log_URL if one doesn't yet exist
    """

    # btype must be the same as helper target in helper.py
    btype = "u_logurl"
    needed_colls = [f"{TARGET_COLL}"]

    def construct_global_ctx(self):
        """Constructs the global context"""
        super().construct_global_ctx()
        gtx = self.gtx
        rc = self.rc
        rc.coll = f"{TARGET_COLL}"
        if not rc.database:
            rc.database = rc.databases[0]["name"]
        gtx[rc.coll] = sorted(all_docs_from_collection(rc.client, rc.coll), key=_id_key)
        gtx["all_docs_from_collection"] = all_docs_from_collection
        gtx["float"] = float
        gtx["str"] = str
        gtx["zip"] = zip

    def db_updater(self):
        rc = self.rc
        key = rc.projectum_id
        filterid = {"_id": key}
        found_projectum = rc.client.find_one(rc.database, rc.coll, filterid)

        # id exists
        if found_projectum is not None:
            rc.client.update_one(rc.database, rc.coll, filterid, {"log_url": rc.log_url})
            print(f"{rc.projectum_id} has been updated with a log_url of {rc.log_url}")
        else:
            # find all similar projectum ids
            pra = fragment_retrieval(self.gtx["projecta"], ["_id"], rc.projectum_id)

            # no matches to id fragment and id does not exist
            if len(pra) == 0:
                raise RuntimeError("Please input a valid projectum id or a valid fragment of a projectum id")

            # id fragment and no inputted number
            elif not rc.index:
                print("There does not seem to be a projectum with this exact name in this database.")
                print("However, there are projecta with similar names: ")
                for i in range(len(pra)):
                    print(f"{i + 1}. {pra[i].get('_id')}     current url: {pra[i].get('log_url')}")
                print(
                    "Please rerun the u_logurl helper with the same name as previously inputted, "
                    "but with the addition of -i followed by a number corresponding to one of the above listed "
                    "projectum ids that you would like to update."
                )

            # id fragment and inputted number
            else:
                if int(rc.index) < 1 or int(rc.index) > len(pra):
                    raise RuntimeError("Sorry, you picked an invalid number.")
                else:
                    filterid = {"_id": pra[int(rc.index) - 1].get("_id")}
                    rc.client.update_one(rc.database, rc.coll, filterid, {"log_url": rc.log_url})
                    print(f"{pra[int(rc.index) - 1].get('_id')} has been updated with a log_url of {rc.log_url}")

        return



import copy
import os
from pathlib import Path

import pytest
import requests_mock

from regolith.main import main
from regolith.schemas import alloweds

MILESTONE_TYPES = alloweds.get("MILESTONE_TYPES")

dash = "-"
helper_map = [
    (
        ["helper", "attestations", "--grant", "dmref15", "--no-plot", "--verbose"],
        "Instructions/Notes:\n"
        "  Quarters are: Q1 July thru Sept, Q2 Oct - Dec, Q3 Jan - Mar, Q4 Apr - Jun\n"
        "  Grad salaries are about $3400 per month\n"
        "Collecting Appointments for grant dmref15:\n"
        "scopatz, from 2019-02-01 to 2019-03-31, loading 0.75. Total months: 1.43\n"
        "\n-----------\nLoadings by month\n------------\n"
        "2018-05-01:\n"
        "2018-06-01:\n"
        "2018-07-01:\n"
        "2018-08-01:\n"
        "2018-09-01:\n"
        "2018-10-01:\n"
        "2018-11-01:\n"
        "2018-12-01:\n"
        "2019-01-01:\n"
        "2019-02-01:\n"
        "    scopatz\tloading: 2550.0\n"
        "2019-03-01:\n"
        "    scopatz\tloading: 2550.0\n"
        "2019-04-01:\n"
        "2019-05-01:\n"
        "\n----------------\nExpenses\n----------------\n"
        "2018-01-10 (reimb date), 2018-01-10 (expense date): amount: 500, \n"
        "2019-02-15 (reimb date), 2018-01-10 (expense date): amount: 1000, \n"
        "  payee: scopatz purpose: testing the databallectionsse\n"
        "2018-05-01: expenses monthly total = 0.00\n"
        "2018-06-01: expenses monthly total = 0.00\n"
        "2018-07-01: expenses monthly total = 0.00\n"
        "2018-08-01: expenses monthly total = 0.00\n"
        "2018-09-01: expenses monthly total = 0.00\n"
        "2018-10-01: expenses monthly total = 0.00\n"
        "2018-11-01: expenses monthly total = 0.00\n"
        "2018-12-01: expenses monthly total = 0.00\n"
        "2019-01-01: expenses monthly total = 0.00\n"
        "2019-02-01: expenses monthly total = 1000.00\n"
        "2019-03-01: expenses monthly total = 0.00\n"
        "2019-04-01: expenses monthly total = 0.00\n"
        "2019-05-01: expenses monthly total = 0.00\n"
        "Total spend = 1500\n",
    ),
    (
        [
            "helper",
            "attestations",
            "--begin-date",
            "2019-01-01",
            "--end-date",
            "2019-05-30",
            "--effort-reporting",
            "--no-plot",
        ],
        ",, 2019-01-01, 2019-02-01, 2019-03-01, 2019-04-01, 2019-05-01\n"
        "scopatz, dmref15, 0.0, 75.0, 75.0, 0.0, 0.0\n",
    ),
    (
        ["helper", "l_progress", "-l", "ascopatz", "--date", "2022-01-09"],
        "\nProgress report for ascopatz, generated 2022-01-09\n"
        "*************************[Orphan Projecta]*************************\n"
        "*************************[Finished Projecta]*************************\n"
        "*************************[Proposed Projecta]*************************\n"
        "*************************[In Progress Projecta]*************************\n"
        "sb_firstprojectum\n"
        "  status: started, begin_date: 2020-04-28, due_date: 2021-05-05\n"
        "  description: My first projectum\n"
        "  log_url: https://docs.google.com/document/d/1YC_wtW5Q\n"
        "  milestones:\n"
        "    due: 2020-05-20, Project lead presentation, type: meeting, status: proposed\n"
        "    objective: lead presents background reading and initial project plan\n"
        "    due: 2020-05-27, planning meeting, type: mergedpr, status: proposed\n"
        "    objective: develop a detailed plan with dates\n",
    ),
    (
        ["helper", "l_progress", "-l", "pliu", "--date", "2022-01-09"],
        "\nProgress report for pliu, generated 2022-01-09\n"
        "*************************[Orphan Projecta]*************************\n"
        "pl_thirdprojectum, status: backburner\n"
        "*************************[Finished Projecta]*************************\n"
        "pl_firstprojectum, grant: None\n"
        "  description: None\n"
        "  finished: 2020-07-27\n"
        "*************************[Proposed Projecta]*************************\n"
        "pl_secondprojectum\n"
        "  status: proposed, begin_date: 2020-07-25, due_date: 2021-08-26\n"
        "  description: None\n"
        "  log_url: None\n"
        "  milestones:\n"
        "    due: 2021-08-03, Milestone, type: None, status: converged\n"
        "    objective: None\n"
        "*************************[In Progress Projecta]*************************\n",
    ),
    (
        ["helper", "l_progress", "-v", "-l", "ascopatz", "--date", "2022-01-09"],
        "\nProgress report for ascopatz, generated 2022-01-09\n"
        "*************************[Orphan Projecta]*************************\n"
        "*************************[Finished Projecta]*************************\n"
        "*************************[Proposed Projecta]*************************\n"
        "*************************[In Progress Projecta]*************************\n"
        "sb_firstprojectum\n"
        "  status: started, begin_date: 2020-04-28, due_date: 2021-05-05\n"
        "  description: My first projectum\n"
        "  log_url: https://docs.google.com/document/d/1YC_wtW5Q\n"
        "  team:\n"
        "    group_members: ascopatz\n"
        "    collaborators: aeinstein, pdirac\n"
        "  deliverable:\n"
        "    audience: beginning grad in chemistry\n"
        "    scope: 1. UCs that are supported or some other scope description if it is software\n"
        "           2. sketch of science story if it is paper\n"
        "    platform: description of how and where the audience will access the deliverable.  Journal if it is a paper\n"
        "  milestones:\n"
        "    2020-05-20: Project lead presentation\n"
        "      objective: lead presents background reading and initial project plan\n"
        "      status: proposed\n"
        "    2020-05-27: planning meeting\n"
        "      objective: develop a detailed plan with dates\n"
        "      status: proposed\n",
    ),
    (
        ["helper", "l_milestones", "--verbose"],
        "2021-08-03 (milest): lead: pliu, pl_secondprojectum, status: converged\n    Type: \n    Title: Milestone\n    log url: None\n    Purpose: None\n    Audience: \n    uuid: milestone_uuid_pl2\n"
        "2021-05-05 (sb_fir): lead: ascopatz, sb_firstprojectum, status: proposed\n    Type: \n    Title: deliverable\n    log url: https://docs.google.com/document/d/1YC_wtW5Q\n    Purpose: deliver\n    Audience: beginning grad in chemistry\n    Notes:\n      - deliverable note\n    uuid: sb_firstprojectum\n"
        "2020-05-27 (milest): lead: ascopatz, sb_firstprojectum, status: proposed\n    Type: mergedpr\n    Title: planning meeting\n    log url: https://docs.google.com/document/d/1YC_wtW5Q\n    Purpose: develop a detailed plan with dates\n    Audience: ascopatz, scopatz, ascopatz\n    uuid: milestone_uuid_sb1_2\n"
        "2020-05-20 (milest): lead: ascopatz, sb_firstprojectum, status: proposed\n    Type: meeting\n    Title: Project lead presentation\n    log url: https://docs.google.com/document/d/1YC_wtW5Q\n    Purpose: lead presents background reading and initial project plan\n    Audience: ascopatz, scopatz, ascopatz\n    Notes:\n      - do background reading\n      - understand math\n    uuid: milestone_uuid_sb1\n",
    ),
    (
        ["helper", "l_milestones", "--verbose", "--current"],
        "2021-08-03 (milest): lead: pliu, pl_secondprojectum, status: converged\n    Type: \n    Title: Milestone\n    log url: None\n    Purpose: None\n    Audience: \n    uuid: milestone_uuid_pl2\n"
        "2021-05-05 (sb_fir): lead: ascopatz, sb_firstprojectum, status: proposed\n    Type: \n    Title: deliverable\n    log url: https://docs.google.com/document/d/1YC_wtW5Q\n    Purpose: deliver\n    Audience: beginning grad in chemistry\n    Notes:\n      - deliverable note\n    uuid: sb_firstprojectum\n"
        "2020-05-27 (milest): lead: ascopatz, sb_firstprojectum, status: proposed\n    Type: mergedpr\n    Title: planning meeting\n    log url: https://docs.google.com/document/d/1YC_wtW5Q\n    Purpose: develop a detailed plan with dates\n    Audience: ascopatz, scopatz, ascopatz\n    uuid: milestone_uuid_sb1_2\n"
        "2020-05-20 (milest): lead: ascopatz, sb_firstprojectum, status: proposed\n    Type: meeting\n    Title: Project lead presentation\n    log url: https://docs.google.com/document/d/1YC_wtW5Q\n    Purpose: lead presents background reading and initial project plan\n    Audience: ascopatz, scopatz, ascopatz\n    Notes:\n      - do background reading\n      - understand math\n    uuid: milestone_uuid_sb1\n",
    ),
    (
        ["helper", "l_milestones", "--verbose", "--current", "--by-prum"],
        f"{dash * 50}\n"
        f"2021-08-03 (milest): lead: pliu, pl_secondprojectum, status: converged\n    Type: \n    Title: Milestone\n    log url: None\n    Purpose: None\n    Audience: \n    uuid: milestone_uuid_pl2\n"
        f"{dash * 50}\n"
        f"2021-05-05 (sb_fir): lead: ascopatz, sb_firstprojectum, status: proposed\n    Type: \n    Title: deliverable\n    log url: https://docs.google.com/document/d/1YC_wtW5Q\n    Purpose: deliver\n    Audience: beginning grad in chemistry\n    Notes:\n      - deliverable note\n    uuid: sb_firstprojectum\n"
        f"2020-05-27 (milest): lead: ascopatz, sb_firstprojectum, status: proposed\n    Type: mergedpr\n    Title: planning meeting\n    log url: https://docs.google.com/document/d/1YC_wtW5Q\n    Purpose: develop a detailed plan with dates\n    Audience: ascopatz, scopatz, ascopatz\n    uuid: milestone_uuid_sb1_2\n"
        f"2020-05-20 (milest): lead: ascopatz, sb_firstprojectum, status: proposed\n    Type: meeting\n    Title: Project lead presentation\n    log url: https://docs.google.com/document/d/1YC_wtW5Q\n    Purpose: lead presents background reading and initial project plan\n    Audience: ascopatz, scopatz, ascopatz\n    Notes:\n      - do background reading\n      - understand math\n    uuid: milestone_uuid_sb1\n",
    ),
    (
        ["helper", "l_milestones", "--verbose", "--all"],
        "2021-08-26 (pl_fir): lead: pliu, pl_firstprojectum, status: finished\n    Type: \n    Title: deliverable\n    log url: None\n    Purpose: deliver\n    Audience: \n    uuid: pl_firstprojectum\n2021-08-26 (pl_sec): lead: pliu, pl_secondprojectum, status: finished\n    Type: \n    Title: deliverable\n    log url: None\n    Purpose: deliver\n    Audience: \n    uuid: pl_secondprojectum\n2021-08-26 (pl_thi): lead: pliu, pl_thirdprojectum, status: finished\n    Type: \n    Title: deliverable\n    log url: None\n    Purpose: deliver\n    Audience: \n    uuid: pl_thirdprojectum\n2021-08-03 (kopl_f): lead: pliu, pl_firstprojectum, status: backburner\n    Type: meeting\n    Title: Kickoff\n    log url: None\n    Purpose: None\n    Audience: \n    uuid: kopl_firstprojectum\n2021-08-03 (milest): lead: pliu, pl_firstprojectum, status: converged\n    Type: \n    Title: Milestone\n    log url: None\n    Purpose: None\n    Audience: \n    uuid: milestone_uuid_pl1\n2021-08-03 (kopl_s): lead: pliu, pl_secondprojectum, status: backburner\n    Type: meeting\n    Title: Kickoff\n    log url: None\n    Purpose: None\n    Audience: \n    uuid: kopl_secondprojectum\n2021-08-03 (milest): lead: pliu, pl_secondprojectum, status: converged\n    Type: \n    Title: Milestone\n    log url: None\n    Purpose: None\n    Audience: \n    uuid: milestone_uuid_pl2\n2021-08-03 (kopl_t): lead: pliu, pl_thirdprojectum, status: backburner\n    Type: meeting\n    Title: Kickoff\n    log url: None\n    Purpose: None\n    Audience: \n    uuid: kopl_thirdprojectum\n2021-08-03 (milest): lead: pliu, pl_thirdprojectum, status: converged\n    Type: \n    Title: Milestone\n    log url: None\n    Purpose: None\n    Audience: \n    uuid: milestone_uuid_pl3\n2021-05-05 (sb_fir): lead: ascopatz, sb_firstprojectum, status: proposed\n    Type: \n    Title: deliverable\n    log url: https://docs.google.com/document/d/1YC_wtW5Q\n    Purpose: deliver\n    Audience: beginning grad in chemistry\n    Notes:\n      - deliverable note\n    uuid: sb_firstprojectum\n2021-05-03 (koab_i): lead: abeing, ab_inactive, status: backburner\n    Type: meeting\n    Title: Kickoff\n    log url: None\n    Purpose: None\n    Audience: \n    uuid: koab_inactive\n2021-05-03 (ab_ina): lead: abeing, ab_inactive, status: paused\n    Type: \n    Title: deliverable\n    log url: None\n    Purpose: deliver\n    Audience: \n    uuid: ab_inactive\n2021-05-03 (milest): lead: abeing, ab_inactive, status: converged\n    Type: \n    Title: Milestone\n    log url: None\n    Purpose: None\n    Audience: \n    uuid: milestone_uuid_inactive\n"
        "2020-05-27 (milest): lead: ascopatz, sb_firstprojectum, status: proposed\n    Type: mergedpr\n    Title: planning meeting\n    log url: https://docs.google.com/document/d/1YC_wtW5Q\n    Purpose: develop a detailed plan with dates\n    Audience: ascopatz, scopatz, ascopatz\n    uuid: milestone_uuid_sb1_2\n"
        "2020-05-20 (milest): lead: ascopatz, sb_firstprojectum, status: proposed\n    Type: meeting\n    Title: Project lead presentation\n    log url: https://docs.google.com/document/d/1YC_wtW5Q\n    Purpose: lead presents background reading and initial project plan\n    Audience: ascopatz, scopatz, ascopatz\n    Notes:\n      - do background reading\n      - understand math\n    uuid: milestone_uuid_sb1\n"
        "2020-05-06 (kosb_f): lead: ascopatz, sb_firstprojectum, status: finished\n    Type: meeting\n    Title: Kick off meeting\n    log url: https://docs.google.com/document/d/1YC_wtW5Q\n    Purpose: introduce project to the lead\n    Audience: ascopatz, scopatz, ascopatz\n    Notes:\n      - kickoff note\n    uuid: kosb_firstprojectum\n",
    ),
    (
        ["helper", "l_milestones", "--verbose", "--person", "aeinstein"],
        "2021-05-05 (sb_fir): lead: ascopatz, sb_firstprojectum, status: proposed\n    Type: \n    Title: deliverable\n    log url: https://docs.google.com/document/d/1YC_wtW5Q\n    Purpose: deliver\n    Audience: beginning grad in chemistry\n    Notes:\n      - deliverable note\n    uuid: sb_firstprojectum\n"
        "2020-05-27 (milest): lead: ascopatz, sb_firstprojectum, status: proposed\n    Type: mergedpr\n    Title: planning meeting\n    log url: https://docs.google.com/document/d/1YC_wtW5Q\n    Purpose: develop a detailed plan with dates\n    Audience: ascopatz, scopatz, ascopatz\n    uuid: milestone_uuid_sb1_2\n"
        "2020-05-20 (milest): lead: ascopatz, sb_firstprojectum, status: proposed\n    Type: meeting\n    Title: Project lead presentation\n    log url: https://docs.google.com/document/d/1YC_wtW5Q\n    Purpose: lead presents background reading and initial project plan\n    Audience: ascopatz, scopatz, ascopatz\n    Notes:\n      - do background reading\n      - understand math\n    uuid: milestone_uuid_sb1\n",
    ),
    (
        ["helper", "l_milestones", "--verbose", "--stati", "finished"],
        "2021-08-26 (pl_sec): lead: pliu, pl_secondprojectum, status: finished\n    Type: \n    Title: deliverable\n    log url: None\n    Purpose: deliver\n    Audience: \n    uuid: pl_secondprojectum\n2020-05-06 (kosb_f): lead: ascopatz, sb_firstprojectum, status: finished\n    Type: meeting\n    Title: Kick off meeting\n    log url: https://docs.google.com/document/d/1YC_wtW5Q\n    Purpose: introduce project to the lead\n    Audience: ascopatz, scopatz, ascopatz\n    Notes:\n      - kickoff note\n    uuid: kosb_firstprojectum\n",
    ),
    (
        ["helper", "l_milestones", "--verbose", "--finished"],
        "2021-08-26 (pl_sec): lead: pliu, pl_secondprojectum, status: finished\n    Type: \n    Title: deliverable\n    log url: None\n    Purpose: deliver\n    Audience: \n    uuid: pl_secondprojectum\n2020-05-06 (kosb_f): lead: ascopatz, sb_firstprojectum, status: finished\n    Type: meeting\n    Title: Kick off meeting\n    log url: https://docs.google.com/document/d/1YC_wtW5Q\n    Purpose: introduce project to the lead\n    Audience: ascopatz, scopatz, ascopatz\n    Notes:\n      - kickoff note\n    uuid: kosb_firstprojectum\n",
    ),
    (
        ["helper", "l_milestones", "--verbose", "--lead", "ascopatz"],
        "2021-05-05 (sb_fir): lead: ascopatz, sb_firstprojectum, status: proposed\n    Type: \n    Title: deliverable\n    log url: https://docs.google.com/document/d/1YC_wtW5Q\n    Purpose: deliver\n    Audience: beginning grad in chemistry\n    Notes:\n      - deliverable note\n    uuid: sb_firstprojectum\n"
        "2020-05-27 (milest): lead: ascopatz, sb_firstprojectum, status: proposed\n    Type: mergedpr\n    Title: planning meeting\n    log url: https://docs.google.com/document/d/1YC_wtW5Q\n    Purpose: develop a detailed plan with dates\n    Audience: ascopatz, scopatz, ascopatz\n    uuid: milestone_uuid_sb1_2\n"
        "2020-05-20 (milest): lead: ascopatz, sb_firstprojectum, status: proposed\n    Type: meeting\n    Title: Project lead presentation\n    log url: https://docs.google.com/document/d/1YC_wtW5Q\n    Purpose: lead presents background reading and initial project plan\n    Audience: ascopatz, scopatz, ascopatz\n    Notes:\n      - do background reading\n      - understand math\n    uuid: milestone_uuid_sb1\n",
    ),
    (
        ["helper", "l_projecta", "--verbose", "--orphan"],
        "ab_inactive\n    status: backburner, begin_date: 2020-05-03, due_date: 2021-05-03, grant: dmref15\n    description: a prum that has various inactive states in milestones and overall\n    team:\n        lead: abeing\n        group_members: None\n        collaborators: None\n",
    ),
    (
        ["helper", "l_projecta", "--verbose", "--lead", "ascopatz"],
        "sb_firstprojectum\n    status: started, begin_date: 2020-04-28, due_date: 2021-05-05, grant: SymPy-1.1\n    description: My first projectum\n    team:\n        lead: ascopatz\n        group_members: ascopatz\n        collaborators: aeinstein, pdirac\n",
    ),
    (
        ["helper", "l_projecta", "--verbose", "--person", "ascopatz"],
        "sb_firstprojectum\n    status: started, begin_date: 2020-04-28, due_date: 2021-05-05, grant: SymPy-1.1\n    description: My first projectum\n    team:\n        lead: ascopatz\n        group_members: ascopatz\n        collaborators: aeinstein, pdirac\n",
    ),
    (["helper", "l_projecta", "--grant", "SymPy-1.1"], "sb_firstprojectum (started)\n"),
    (
        ["helper", "l_projecta", "--grp_by_lead"],
        "abeing:\n    ab_inactive (backburner)\npliu:\n    pl_firstprojectum (finished)\n    pl_secondprojectum (proposed)\n    pl_thirdprojectum (backburner)\nascopatz:\n    sb_firstprojectum (started)\n",
    ),
    (
        ["helper", "l_projecta", "--all"],
        "ab_inactive (backburner)\npl_firstprojectum (finished)\npl_secondprojectum (proposed)\npl_thirdprojectum (backburner)\nsb_firstprojectum (started)\n",
    ),
    (["helper", "l_projecta", "--current"], "pl_secondprojectum (proposed)\nsb_firstprojectum (started)\n"),
    (["helper", "l_projecta", "--grp_by_lead", "-l", "ascopatz"], "ascopatz:\n    sb_firstprojectum (started)\n"),
    (
        ["helper", "l_projecta", "--verbose"],
        "ab_inactive\n    status: backburner, begin_date: 2020-05-03, due_date: 2021-05-03, grant: dmref15\n    description: a prum that has various inactive states in milestones and overall\n    team:\n        lead: abeing\n        group_members: None\n        collaborators: None\n"
        "pl_firstprojectum\n    status: finished, begin_date: 2020-07-25, due_date: 2021-08-26, end_date: 2020-07-27, grant: None\n    description: None\n    team:\n        lead: pliu\n        group_members: None\n        collaborators: None\n"
        "pl_secondprojectum\n    status: proposed, begin_date: 2020-07-25, due_date: 2021-08-26, grant: None\n    description: None\n    team:\n        lead: pliu\n        group_members: None\n        collaborators: None\n"
        "pl_thirdprojectum\n    status: backburner, begin_date: 2020-07-25, due_date: 2021-08-26, grant: None\n    description: None\n    team:\n        lead: pliu\n        group_members: None\n        collaborators: None\n"
        "sb_firstprojectum\n    status: started, begin_date: 2020-04-28, due_date: 2021-05-05, grant: SymPy-1.1\n    description: My first projectum\n    team:\n        lead: ascopatz\n        group_members: ascopatz\n        collaborators: aeinstein, pdirac\n",
    ),
    (
        ["helper", "l_projecta", "--ended", "--date", "2020-06-02"],
        "\nNo projecta finished within the 7 days leading up to 2020-06-02\n",
    ),
    (
        ["helper", "l_grants", "--current", "--date", "2020-05-25"],
        "\nAdministered by: apam\n"
        "  sym2.0         \t awardnr: NF-1234        \t acctn: n/a                 \t 2019-06-01 to 2030-12-31\n",
    ),
    (
        ["helper", "l_grants", "--current", "--date", "2020-05-25", "--reveal-hidden"],
        "\nAdministered by: apam\n"
        "  sym2.0         \t awardnr: NF-1234        \t acctn: n/a                 \t 2019-06-01 to 2030-12-31\n",
    ),
    (
        ["helper", "l_grants", "--current", "--date", "2020-05-25", "--verbose"],
        "\nAdministered by: apam\n"
        "  sym2.0         \t awardnr: NF-1234        \t acctn: n/a                 \t 2019-06-01 to 2030-12-31\n"
        "      funds available: $2,100 on 2021-01-03\n",
    ),
    (
        ["helper", "l_members", "--current", "-v"],
        "    -- Assistant Scientists --\n"
        "Simon J. L. Billinge, professor\n"
        "    email: sb2896@columbia.edu | group_id: sbillinge\n"
        "    github_id: None | orcid: 0000-0002-9432-4248\n"
        "    current organization: The University of South Carolina\n"
        "    current position: Assistant Professor\n",
    ),
    (
        ["helper", "l_members", "-v"],
        "    -- Assistant Scientists --\n"
        "Simon J. L. Billinge, professor\n"
        "    email: sb2896@columbia.edu | group_id: sbillinge\n"
        "    github_id: None | orcid: 0000-0002-9432-4248\n"
        "    current organization: The University of South Carolina\n"
        "    current position: Assistant Professor\n"
        "Anthony Scopatz, professor\n"
        "    email: scopatz@cec.sc.edu | group_id: scopatz\n"
        "    github_id: ascopatz | orcid: 0000-0002-9432-4248\n"
        "    current organization: The University of South Carolina\n"
        "    current position: Assistant Professor\n"
        "    -- Undergrads --\n"
        "Abstract Being, intern\n"
        "    email: None | group_id: abeing\n"
        "    github_id: None | orcid: None\n"
        "    current organization: The University of South Carolina\n"
        "    current position: Assistant Professor\n",
    ),
    (
        ["helper", "l_members", "--prior", "-v"],
        "    -- Assistant Scientists --\n"
        "Anthony Scopatz, professor\n"
        "    email: scopatz@cec.sc.edu | group_id: scopatz\n"
        "    github_id: ascopatz | orcid: 0000-0002-9432-4248\n"
        "    current organization: The University of South Carolina\n"
        "    current position: Assistant Professor\n"
        "    -- Undergrads --\n"
        "Abstract Being, intern\n"
        "    email: None | group_id: abeing\n"
        "    github_id: None | orcid: None\n"
        "    current organization: The University of South Carolina\n"
        "    current position: Assistant Professor\n",
    ),
    (["helper", "l_members", "--filter", "name", "sco"], "scopatz    \n"),
    (
        ["helper", "l_members", "--filter", "name", "sco", "-v"],
        "Anthony Scopatz, professor | group_id: scopatz\n"
        "    orcid: 0000-0002-9432-4248 | github_id: ascopatz\n",
    ),
    (
        [
            "helper",
            "l_contacts",
            "run",
            "--name",
            "ny",
            "--inst",
            "col",
            "--notes",
            "coffee",
            "--date",
            "2020-01-15",
            "--range",
            "2",
        ],
        "Anthony B Friend  |  afriend  |  institution: Columbia University  |  email: friend@deed.com\n",
    ),
    (
        [
            "helper",
            "l_contacts",
            "run",
            "--name",
            "ny",
            "--inst",
            "col",
            "--notes",
            "coffee",
            "--date",
            "2019-01-15",
            "--range",
            "2",
        ],
        "\n",
    ),
    (
        ["helper", "l_contacts", "run", "--verbose"],
        "Anthony B Friend\n"
        "    _id: afriend\n"
        "    email: friend@deed.com\n"
        "    institution: Columbia University\n"
        "    department: physics\n"
        "    notes:\n"
        "        -The guy I meet for coffee sometimes\n"
        "    aka:\n"
        "        -A. B. Friend\n        -AB Friend\n        -Tony Friend\n",
    ),
    (
        ["helper", "l_abstract", "--year", "2018", "--author", "afriend"],
        "\n---------------------------------------\n"
        "2018-05-22 - Meeting to check flexibility on dates, Upton NY\n"
        "---------------------------------------\n"
        "Title: Graphitic Dephenestration\n\n"
        "Anthony Scopatz, Anthony B Friend\n\n"
        "Abstract: We pulled apart graphite with tape\n",
    ),
    (
        ["helper", "l_abstract", "--year", "2018", "--title", "nanostructure"],
        "\n---------------------------------------\n"
        "2018-05-22 - Colloquium Department of Physics, Columbia University\n"
        "---------------------------------------\n"
        "Title: Nanostructure challenges and successes from 16th Century warships to 21st Century energy\n\n"
        "Anthony Scopatz\n\n"
        "Abstract: We made the case for local structure\n",
    ),
    (
        ["helper", "l_abstract", "--title", "graphitic"],
        "\n---------------------------------------\n"
        "2018-05-22 - Meeting to check flexibility on dates, Upton NY\n"
        "---------------------------------------\n"
        "Title: Graphitic Dephenestration\n\n"
        "Anthony Scopatz, Anthony B Friend\n\n"
        "Abstract: We pulled apart graphite with tape\n",
    ),
    (
        ["helper", "l_abstract", "--title", "graphitic", "--loc-inst", "upton"],
        "\n---------------------------------------\n"
        "2018-05-22 - Meeting to check flexibility on dates, Upton NY\n"
        "---------------------------------------\n"
        "Title: Graphitic Dephenestration\n\n"
        "Anthony Scopatz, Anthony B Friend\n\n"
        "Abstract: We pulled apart graphite with tape\n",
    ),
    (
        ["helper", "l_abstract", "--loc-inst", "upton"],
        "\n---------------------------------------\n"
        "2018-05-22 - 2018 NSLS-II and CFN Users Meeting, Upton NY\n"
        "---------------------------------------\n"
        "Title: ClusterMining: extracting core structures of metallic nanoparticles from the atomic pair distribution function\n\n"
        "Anthony Scopatz\n\n"
        "Abstract: We pulled apart graphite with tape\n"
        "\n---------------------------------------\n"
        "2018-05-22 - Meeting to check flexibility on dates, Upton NY\n"
        "---------------------------------------\n"
        "Title: Graphitic Dephenestration\n\n"
        "Anthony Scopatz, Anthony B Friend\n\n"
        "Abstract: We pulled apart graphite with tape\n",
    ),
    (
        ["helper", "l_abstract", "--loc-inst", "upton", "--year", "2018"],
        "\n---------------------------------------\n"
        "2018-05-22 - 2018 NSLS-II and CFN Users Meeting, Upton NY\n"
        "---------------------------------------\n"
        "Title: ClusterMining: extracting core structures of metallic nanoparticles from the atomic pair distribution function\n\n"
        "Anthony Scopatz\n\n"
        "Abstract: We pulled apart graphite with tape\n"
        "\n---------------------------------------\n"
        "2018-05-22 - Meeting to check flexibility on dates, Upton NY\n"
        "---------------------------------------\n"
        "Title: Graphitic Dephenestration\n\n"
        "Anthony Scopatz, Anthony B Friend\n\n"
        "Abstract: We pulled apart graphite with tape\n",
    ),
    (
        ["helper", "l_abstract", "--year", "2018"],
        "\n---------------------------------------\n"
        "2018-05-22 - Colloquium Department of Physics, Columbia University\n"
        "---------------------------------------\n"
        "Title: Nanostructure challenges and successes from 16th Century warships to 21st Century energy\n\n"
        "Anthony Scopatz\n\n"
        "Abstract: We made the case for local structure\n"
        "\n---------------------------------------\n"
        "2018-05-22 - 2018 NSLS-II and CFN Users Meeting, Upton NY\n"
        "---------------------------------------\n"
        "Title: ClusterMining: extracting core structures of metallic nanoparticles from the atomic pair distribution function\n\n"
        "Anthony Scopatz\n\n"
        "Abstract: We pulled apart graphite with tape\n"
        "\n---------------------------------------\n"
        "2018-05-22 - Meeting to check flexibility on dates, Upton NY\n"
        "---------------------------------------\n"
        "Title: Graphitic Dephenestration\n\n"
        "Anthony Scopatz, Anthony B Friend\n\n"
        "Abstract: We pulled apart graphite with tape\n",
    ),
    (
        ["helper", "l_abstract", "--author", "scopatz"],
        "\n---------------------------------------\n"
        "2018-05-22 - Colloquium Department of Physics, Columbia University\n"
        "---------------------------------------\n"
        "Title: Nanostructure challenges and successes from 16th Century warships to 21st Century energy\n\n"
        "Anthony Scopatz\n\n"
        "Abstract: We made the case for local structure\n"
        "\n---------------------------------------\n"
        "2018-05-22 - 2018 NSLS-II and CFN Users Meeting, Upton NY\n"
        "---------------------------------------\n"
        "Title: ClusterMining: extracting core structures of metallic nanoparticles from the atomic pair distribution function\n\n"
        "Anthony Scopatz\n\n"
        "Abstract: We pulled apart graphite with tape\n"
        "\n---------------------------------------\n"
        "2018-05-22 - Meeting to check flexibility on dates, Upton NY\n"
        "---------------------------------------\n"
        "Title: Graphitic Dephenestration\n\n"
        "Anthony Scopatz, Anthony B Friend\n\n"
        "Abstract: We pulled apart graphite with tape\n",
    ),
    (
        ["helper", "l_abstract", "--loc-inst", "upton", "--year", "2018"],
        "\n---------------------------------------\n"
        "2018-05-22 - 2018 NSLS-II and CFN Users Meeting, Upton NY\n"
        "---------------------------------------\n"
        "Title: ClusterMining: extracting core structures of metallic nanoparticles from the atomic pair distribution function\n\n"
        "Anthony Scopatz\n\n"
        "Abstract: We pulled apart graphite with tape\n"
        "\n---------------------------------------\n"
        "2018-05-22 - Meeting to check flexibility on dates, Upton NY\n"
        "---------------------------------------\n"
        "Title: Graphitic Dephenestration\n\n"
        "Anthony Scopatz, Anthony B Friend\n\n"
        "Abstract: We pulled apart graphite with tape\n",
    ),
    (
        ["helper", "l_abstract", "--author", "scopatz", "--loc-inst", "upton"],
        "\n---------------------------------------\n"
        "2018-05-22 - 2018 NSLS-II and CFN Users Meeting, Upton NY\n"
        "---------------------------------------\n"
        "Title: ClusterMining: extracting core structures of metallic nanoparticles from the atomic pair distribution function\n"
        "\nAnthony Scopatz\n"
        "\nAbstract: We pulled apart graphite with tape\n"
        "\n---------------------------------------\n"
        "2018-05-22 - Meeting to check flexibility on dates, Upton NY\n"
        "---------------------------------------\n"
        "Title: Graphitic Dephenestration\n\n"
        "Anthony Scopatz, Anthony B Friend\n\n"
        "Abstract: We pulled apart graphite with tape\n",
    ),
    (
        [
            "helper",
            "l_abstract",
            "--author",
            "scopatz",
            "--year",
            "2018",
            "--loc-inst",
            "upton",
            "--title",
            "graphitic",
        ],
        "\n---------------------------------------\n"
        "2018-05-22 - Meeting to check flexibility on dates, Upton NY\n"
        "---------------------------------------\n"
        "Title: Graphitic Dephenestration\n\n"
        "Anthony Scopatz, Anthony B Friend\n\n"
        "Abstract: We pulled apart graphite with tape\n",
    ),
    (
        ["helper", "l_abstract", "--loc-inst", "columbiau"],
        "\n---------------------------------------\n"
        "2018-05-22 - Colloquium Department of Physics, Columbia University\n"
        "---------------------------------------\n"
        "Title: Nanostructure challenges and successes from 16th Century warships to 21st Century energy\n\n"
        "Anthony Scopatz\n\n"
        "Abstract: We made the case for local structure\n",
    ),
    (
        ["helper", "l_todo", "--assigned-to", "sbillinge", "--date", "2020-05-01"],
        "If the indices are far from being in numerical order, please renumber them by running regolith helper u_todo -r\n"
        "(index) action (days to due date|importance|expected duration (mins)|tags|assigned by)\n"
        "--------------------------------------------------------------------------------\n"
        "started:\n"
        "(0)(89 days): (2) prepare the presentation (89|0|30.0|downtime|sbillinge|2saefa)\n"
        "     - about 10 minutes\n"
        "     - don't forget to upload to the website\n"
        "(2)(79 days): (1) read paper (79|2|60.0|reading,downtime|scopatz|1saefa)\n"
        "----------------------------------------------------------------------------\n"
        "(importance)(days to due): (Task number) Task (decreasing priority going up)\n"
        "----------------------------------------------------------------------------\n"
        "2020-07-19(79 days): (1) read paper (79|2|60.0|reading,downtime|scopatz|1saefa)\n"
        "------------------------------\n"
        "Deadlines:\n"
        "------------------------------\n",
    ),
    (
        [
            "helper",
            "l_todo",
            "--short",
            "65",
            "--date",
            "2020-07-13",
            "--assigned-by",
            "scopatz",
            "--assigned-to",
            "sbillinge",
        ],
        "If the indices are far from being in numerical order, please renumber them by running regolith helper u_todo -r\n"
        "(index) action (days to due date|importance|expected duration (mins)|tags|assigned by)\n"
        "--------------------------------------------------------------------------------\n"
        "started:\n"
        "(2)(6 days): (1) read paper (6|2|60.0|reading,downtime|scopatz|1saefa)\n"
        "----------------------------------------------------------------------------\n"
        "(importance)(days to due): (Task number) Task (decreasing priority going up)\n"
        "----------------------------------------------------------------------------\n"
        "2020-07-19(6 days): (1) read paper (6|2|60.0|reading,downtime|scopatz|1saefa)\n"
        "------------------------------\n"
        "Deadlines:\n"
        "------------------------------\n",
    ),
    (
        [
            "helper",
            "l_todo",
            "--tags",
            "downtime",
            "--date",
            "2020-07-13",
            "--assigned-by",
            "sbillinge",
            "--assigned-to",
            "sbillinge",
        ],
        "If the indices are far from being in numerical order, please renumber them by running regolith helper u_todo -r\n"
        "(index) action (days to due date|importance|expected duration (mins)|tags|assigned by)\n"
        "--------------------------------------------------------------------------------\n"
        "started:\n"
        "(0)(16 days): (2) prepare the presentation (16|0|30.0|downtime|sbillinge|2saefa)\n"
        "     - about 10 minutes\n"
        "     - don't forget to upload to the website\n"
        "----------------------------------------------------------------------------\n"
        "(importance)(days to due): (Task number) Task (decreasing priority going up)\n"
        "----------------------------------------------------------------------------\n"
        "------------------------------\n"
        "Deadlines:\n"
        "------------------------------\n",
    ),
    (["helper", "l_todo", "--assigned-to", "wrong_id"], "The id you entered can't be found in todos.yml.\n"),
    (
        ["helper", "l_todo", "-o", "--date", "2021-4-10", "--assigned-to", "sbillinge", "--short"],
        "If the indices are far from being in numerical order, please renumber them by running regolith helper u_todo -r\n"
        "(index) action (days to due date|importance|expected duration (mins)|tags|assigned by)\n"
        "--------------------------------------------------------------------------------\n"
        "started:\n"
        "(0)(-255 days): (2) prepare the presentation (-255|0|30.0|downtime|sbillinge|2saefa)\n"
        "     - about 10 minutes\n"
        "     - don't forget to upload to the website\n"
        "----------------------------------------------------------------------------\n"
        "(importance)(days to due): (Task number) Task (decreasing priority going up)\n"
        "----------------------------------------------------------------------------\n"
        "------------------------------\n"
        "Deadlines:\n"
        "------------------------------\n"
        "------------------------------\n"
        "Outstanding Reviews:\n"
        "------------------------------\n"
        "accepted:\n"
        "Manuscript by Wingit in Nature is due on 2021-04-11\n",
    ),
    (
        ["helper", "l_currentappointments", "-d", "2021-08-10"],
        "scopatz future_grant n/a 1.0 2020-09-01 2021-08-31\n",
    ),
    (["helper", "l_currentappointments", "-d", "2020-06-01"], "scopatz abc42 abc42 0.8 2020-06-01 2020-08-31\n"),
    (["helper", "l_currentappointments", "-d", "2020-01-01", "-s"], "scopatz sym sym 1.0 2020-01-01 2020-05-15\n"),
    (["helper", "v_meetings", "--test"], "Meeting validator helper\n"),
    (
        ["helper", "l_reimbstatus", "scopatz"],
        "Reimbursed expenses:\n"
        "\n"
        "Submitted expenses:\n"
        " - 180110 - testing the databallectionsse 2018-01-01 to 2018-01-10,\n"
        "   Expenses: unseg=550.00, Seg=0.00, Total=550.00, Where: Columbia, When: tbd\n"
        " - 180110 - testing the databallectionsse 2018-01-01 to 2018-01-10,\n"
        "   Expenses: unseg=550.00, Seg=0.00, Total=550.00, Where: Columbia, When: 2019-09-05\n"
        "   Grants: dmref15, SymPy-1.1\n"
        "this expense was used to get the work done\n"
        "\nUnsubmitted expenses:\n"
        "\nFuture expenses:\n",
    ),
    (
        ["helper", "l_reimbstatus", "sbillinge"],
        "Reimbursed expenses:\n"
        " - 190110 - testing 2019-01-01 to 2019-01-10,\n"
        "   Requested: 10, Reimbursed: 100, Date: 2019-09-15, Grants: SymPy-1.1\n"
        "\nSubmitted expenses:\n"
        "\nUnsubmitted expenses:\n"
        "\nFuture expenses:\n"
        "\nThese expenses have invalid statuses:\n"
        "test3\n",
    ),
    # these updaters are really listers because they list the current state
    (
        [
            "helper",
            "a_projectum",
            "New projectum",
            "lyang",
            "--date",
            "2020-04-29",
            "--collaborators",
            "afriend",
            "--description",
            "more work",
            "--group-members",
            "ascopatz",
            "--grants",
            "SymPy-1.1",
            "--due-date",
            "2021-01-01",
            "--notes",
            "new note",
        ],
        "ly_newprojectum has been added in projecta\n",
    ),
    (
        [
            "helper",
            "u_milestone",
            "--milestone_uuid",
            "kosb_fir",
            "--name",
            "Kick off meeting",
            "--date",
            "2020-05-07",
            "--objective",
            "introduce project to the lead",
            "--audience",
            "lead",
            "pi",
            "group_members",
            "--status",
            "converged",
            "--due-date",
            "2020-06-01",
            "--notes",
            "do this",
            "do that",
            "--type",
            "meeting",
            "--finish",
        ],
        "The milestone 'Kick off meeting' has been marked as finished in prum sb_firstprojectum.\n",
    ),
    (
        ["helper", "u_milestone", "--milestone_uuid", "bad_id"],
        "Failed to update projecta.\n"
        "No ids were found that match your milestone_uuid entry (bad_id).\n"
        "Make sure you have entered the correct uuid or uuid fragment and rerun the helper.\n\n",
    ),
    (
        [
            "helper",
            "u_milestone",
            "--milestone_uuid",
            "pl",
            "--status",
            "finished",
            "--due-date",
            "2023-01-01",
            "--notes",
            "do this",
            "do that",
            "--type",
            "mergedpr",
        ],
        "Failed to update projecta.\n"
        "Multiple ids match your milestone_uuid entry (pl).\n"
        "Try entering more characters of the uuid and rerun the helper.\n\n",
    ),
    (
        [
            "helper",
            "u_milestone",
            "--projectum_id",
            "pl",
            "--name",
            "new milestone",
            "--due_date",
            "2023-01-01",
            "--objective",
            "do all the things to complete this milestone",
            "--notes",
            "do this",
            "do that",
            "--type",
            "mergedpr",
        ],
        "Projectum not found. Projecta with similar names: \n"
        "pl_firstprojectum\n"
        "pl_secondprojectum\n"
        "pl_thirdprojectum\n"
        "Please rerun the helper specifying the complete ID.\n"
        "If your prum id looks correct, check that this id is in the collection "
        "in the database test.\n"
        "If this is not the case, rerun with --database set to the database where the item is located.\n",
    ),
    (
        ["helper", "u_logurl", "sb", "--index", "1", "https://docs.google.com/document/d/1pQMFpuI"],
        "sb_firstprojectum has been updated with a log_url of https://docs.google.com/document/d/1pQMFpuI\n",
    ),
    (
        ["helper", "u_logurl", "ly", "https://docs.google.com/document/d/1pQMFpuI"],
        "There does not seem to be a projectum with this exact name in this database.\n"
        "However, there are projecta with similar names: \n"
        "1. ly_newprojectum     current url: \n"
        "Please rerun the u_logurl helper with the same name as previously inputted, "
        "but with the addition of -i followed by a number corresponding to one of the "
        "above listed projectum ids that you would like to update.\n",
    ),
    (
        [
            "helper",
            "u_contact",
            "afriend",
            "--index",
            "2",
            "--notes",
            "Test note",
            "--aliases",
            "Friend",
            "--date",
            "2020-01-02",
        ],
        "afriend has been added/updated in contacts\n",
    ),
    (
        ["helper", "u_contact", "Anthony", "--date", "2020-01-02"],
        "Please rerun the helper by hitting up arrow and adding '-i list-index' to "
        "update the list item 'list-index', e.g., 'regolith helper eins -i 2'. For "
        "new contacts --name (-n) and --institution (-o) are required:\n"
        "1. Anthony as a new contact\n"
        "2. Anthony B Friend\n"
        "   id: afriend\n"
        "   email: friend@deed.com\n"
        "   institution: columbiau\n"
        "   department: physics\n"
        "   notes: ['The guy I meet for coffee sometimes', 'Test note']\n"
        "   aliases: ['A. B. Friend', 'AB Friend', 'Tony Friend', 'Friend']\n",
    ),
    (
        ["helper", "u_contact", "Maria", "--date", "2020-01-02"],
        "Please rerun the helper by hitting up arrow and adding '-i list-index' to "
        "update the list item 'list-index', e.g., 'regolith helper eins -i 2'. For "
        "new contacts --name (-n) and --institution (-o) are required:\n"
        "1. Maria as a new contact\n",
    ),
    (
        [
            "helper",
            "a_todo",
            "test a_todo",
            "6",
            "50",
            "--assigned-to",
            "sbillinge",
            "--assigned-by",
            "sbillinge",
            "--begin-date",
            "2020-07-06",
            "--importance",
            "2",
            "--deadline",
            "--notes",
            "test notes 1",
            "test notes 2",
            "--tags",
            "tag1",
            "tag2",
            "--date",
            "2020-07-10",
            "--milestone_uuid",
            "milestone_uuid_sb1_2",
        ],
        "The milestone uuid milestone_uuid_sb1_2 in projectum sb_firstprojectum has been updated in projecta.\n"
        'The task "test a_todo" for sbillinge has been added in todos collection.\n',
    ),
    (
        [
            "helper",
            "f_todo",
            "--index",
            "3",
            "--assigned-to",
            "sbillinge",
            "--end-date",
            "2020-07-20",
            "--date",
            "2020-07-13",
        ],
        'The task "(3) test a_todo" in test for sbillinge has been marked as finished.\n',
    ),
    (
        ["helper", "f_todo", "--assigned-to", "sbillinge", "--date", "2020-07-13"],
        "If the indices are far from being in numerical order, please renumber them by running regolith helper u_todo -r\n"
        "Please choose from one of the following to update:\n"
        "(index) action (days to due date|importance|expected duration (mins)|tags|assigned by)\n"
        "--------------------------------------------------------------------------------\n"
        "started:\n"
        "(0)(16 days): (2) prepare the presentation (16|0|30.0|downtime|sbillinge|2saefa)\n"
        "     - about 10 minutes\n"
        "     - don't forget to upload to the website\n"
        "(2)(6 days): (1) read paper (6|2|60.0|reading,downtime|scopatz|1saefa)\n"
        "----------------------------------------------------------------------------\n"
        "(importance)(days to due): (Task number) Task (decreasing priority going up)\n"
        "----------------------------------------------------------------------------\n"
        "2020-07-19(6 days): (1) read paper (6|2|60.0|reading,downtime|scopatz|1saefa)\n"
        "------------------------------\n"
        "Deadlines:\n"
        "------------------------------\n",
    ),
    (
        ["helper", "f_todo", "--index", "99100"],
        "WARNING: indices >= 9900 are used for milestones which should be finished using u_milestone and not f_todo\n",
    ),
    (
        [
            "helper",
            "u_todo",
            "--index",
            "3",
            "--assigned-to",
            "sbillinge",
            "--description",
            "update the description",
            "--due-date",
            "2020-07-06",
            "--estimated-duration",
            "35",
            "--importance",
            "2",
            "--status",
            "finished",
            "--notes",
            "some new notes",
            "notes2",
            "--tags",
            "newtag1",
            "newtag2",
            "--begin-date",
            "2020-06-06",
            "--deadline",
            "t",
            "--end-date",
            "2020-07-07",
            "--date",
            "2020-07-13",
        ],
        'The task "(3) test a_todo" in test for sbillinge has been updated.\n',
    ),
    (
        [
            "helper",
            "u_todo",
            "--assigned-to",
            "sbillinge",
            "--stati",
            "started",
            "finished",
            "--filter",
            "description",
            "the",
            "--date",
            "2020-07-13",
        ],
        "If the indices are far from being in numerical order, please renumber them by running regolith helper u_todo -r\n"
        "Please choose from one of the following to update:\n"
        "(index) action (days to due date|importance|expected duration (mins)|assigned by)\n"
        "--------------------------------------------------------------------------------\n"
        "started:\n"
        "(0)(16 days): (2) prepare the presentation (16|0|30.0|downtime|sbillinge|2saefa)\n"
        "     - about 10 minutes\n"
        "     - don't forget to upload to the website\n"
        "finished:\n"
        "(2)(-7 days): (3) update the description (-7|2|35.0|tag1,tag2,newtag1,newtag2|sbillinge|test-u)\n"
        "     - test notes 1\n"
        "     - test notes 2\n"
        "     - some new notes\n"
        "     - notes2\n"
        "----------------------------------------------------------------------------\n"
        "(importance)(days to due): (Task number) Task (decreasing priority going up)\n"
        "----------------------------------------------------------------------------\n"
        "2020-07-06(-7 days): (3) update the description (-7|2|35.0|tag1,tag2,newtag1,newtag2|sbillinge|test-u)\n"
        "     - test notes 1\n"
        "     - test notes 2\n"
        "     - some new notes\n"
        "     - notes2\n"
        "------------------------------\n"
        "Deadlines:\n"
        "------------------------------\n",
    ),
    (
        ["helper", "u_todo", "--index", "99100"],
        "WARNING: indices >= 9900 are used for milestones which should be updated using u_milestone and not u_todo\n",
    ),
    (
        ["helper", "f_prum", "sb_firstprojectum", "--end-date", "2020-07-01"],
        "sb_firstprojectum status has been updated to finished\n",
    ),
    (
        ["helper", "f_prum", "sb_"],
        "Projectum not found. Projecta with similar names: \n"
        "sb_firstprojectum     status:finished\n"
        "Please rerun the helper specifying the complete ID.\n",
    ),
    (["helper", "lister", "people"], "Results of your search:\nabeing    \nsbillinge    \nscopatz\n"),
    (["helper", "lister", "people", "--kv-filter", "name", "simon"], "Results of your search:\n" "sbillinge\n"),
    (
        ["helper", "lister", "people", "--kv-filter", "name", "simon", "--return-fields", "name", "position"],
        "Results of your search:\nsbillinge    name: Simon J. L. Billinge    position: professor\n",
    ),
    (
        ["helper", "lister", "people", "--keys"],
        "Available keys:\n"
        "['_id', 'active', 'activities', 'aka', 'appointments', 'avatar', 'bio', 'bios', "
        "'committees', 'education', 'email', 'employment', 'facilities', 'funding', "
        "'github_id', 'google_scholar_url', 'grp_mtg_active', 'hindex', "
        "'home_address', 'initials', 'linkedin_url', "
        "'membership', 'miscellaneous', 'name', 'office', 'orcid_id', 'position', "
        "'publicity', 'research_focus_areas', 'service', 'skills', 'teaching', "
        "'title']\n",
    ),
    (
        ["helper", "lister", "people", "--kv-filter", "name", "simon", "--keys"],
        "Results of your search:\nsbillinge\n"
        "Available keys:\n"
        "['_id', 'active', 'activities', 'aka', 'appointments', 'avatar', 'bio', 'bios', "
        "'committees', 'education', 'email', 'employment', 'facilities', 'funding', "
        "'github_id', 'google_scholar_url', 'grp_mtg_active', 'hindex', "
        "'home_address', 'initials', 'linkedin_url', "
        "'membership', 'miscellaneous', 'name', 'office', 'orcid_id', 'position', "
        "'publicity', 'research_focus_areas', 'service', 'skills', 'teaching', "
        "'title']\n",
    ),
    (
        ["helper", "lister", "people", "--kv-filter", "name", "simon", "position", "singer"],
        "There are no results that match your search.\n",
    ),
    (
        [
            "helper",
            "u_institution",
            "columbiau",
            "--aka",
            "ucolumbia",
            "Columbia University in the City of New York",
            "--dept-id",
            "mathematics",
            "--dept-name",
            "Department of Mathematics",
            "--dept-aka",
            "dept. of mathematics",
            "math department",
            "--school-id",
            "cc",
            "--school-name",
            "Columbia College",
            "--school-aka",
            "CC",
            "--date",
            "2020-01-01",
        ],
        "columbiau has been updated/added in institutions\n",
    ),
    (
        ["helper", "u_institution", "col"],
        "Please rerun the helper specifying '-n list-index' to update item number 'list-index':\n"
        "1. col as a new institution.\n"
        "2. columbiau      Columbia University.\n",
    ),
    (
        ["helper", "makeappointments", "run", "--no-gui", "--projection-from-date", "2020-08-31"],
        "WARNING: appointment gap for scopatz from 2019-02-01 to 2019-12-31\n"
        "WARNING: appointment gap for scopatz from 2020-05-16 to 2020-08-31\n"
        "appointments on outdated grants:\n"
        "    person: scopatz, appointment: s20, grant: sym,\n"
        "            from 2020-01-01 until 2020-05-15\n"
        "appointments on depleted grants:\n"
        "    person: scopatz, appointment: f19, grant: dmref15,\n"
        "            from 2019-02-07 until 2019-03-31\n"
        "    person: scopatz, appointment: ss20, grant: abc42,\n"
        "            from 2020-07-15 until 2020-08-31\n"
        "    person: scopatz, appointment: ss21, grant: future_grant,\n"
        "            from 2020-09-06 until 2021-08-31\n"
        "underspent grants:\n"
        "    dmref15: end: 2019-05-01\n"
        "      projected underspend: 54.55 months, balance as of 2020-08-31: 0\n"
        "      required ss+gra burn: -3.41\n"
        "    sym: end: 2030-12-31\n"
        "      projected underspend: 8.0 months, balance as of 2020-08-31: 0\n"
        "      required ss+gra burn: 0.06\n"
        "cumulative underspend = 62.55 months, cumulative months to support = 0\n"
        "overspent grants:\n"
        "    end: 2020-12-31, grant: abc42, overspend amount: -1.41 months\n"
        "    end: 2026-08-30, grant: future_grant, overspend amount: -11.97 months\n"
        "plotting mode is on\n",
    ),
    (
        ["helper", "makeappointments", "run", "--no-gui", "--projection-from-date", "2020-08-31", "-v"],
        "WARNING: appointment gap for scopatz from 2019-02-01 to 2019-12-31\n"
        "WARNING: appointment gap for scopatz from 2020-05-16 to 2020-08-31\n"
        "appointments on outdated grants:\n"
        "    person: scopatz, appointment: s20, grant: sym,\n"
        "            from 2020-01-01 until 2020-05-15\n"
        "appointments on depleted grants:\n"
        "    person: scopatz, appointment: f19, grant: dmref15,\n"
        "            from 2019-02-07 until 2019-03-31\n"
        "    person: scopatz, appointment: ss20, grant: abc42,\n"
        "            from 2020-07-15 until 2020-08-31\n"
        "    person: scopatz, appointment: ss21, grant: future_grant,\n"
        "            from 2020-09-06 until 2021-08-31\n"
        "underspent grants:\n"
        "    dmref15: end: 2019-05-01\n"
        "      projected underspend: 54.55 months, balance as of 2020-08-31: 0\n"
        "      required ss+gra burn: -3.41\n"
        "    sym: end: 2030-12-31\n"
        "      projected underspend: 8.0 months, balance as of 2020-08-31: 0\n"
        "      required ss+gra burn: 0.06\n"
        "cumulative underspend = 62.55 months, cumulative months to support = 0\n"
        "overspent grants:\n"
        "    end: 2020-12-31, grant: abc42, overspend amount: -1.41 months\n"
        "    end: 2026-08-30, grant: future_grant, overspend amount: -11.97 months\n"
        "plotting mode is on\n",
    ),
    (
        ["helper", "makeappointments", "run", "--no-plot", "--projection-from-date", "2020-08-31"],
        "WARNING: appointment gap for scopatz from 2019-02-01 to 2019-12-31\n"
        "WARNING: appointment gap for scopatz from 2020-05-16 to 2020-08-31\n"
        "appointments on outdated grants:\n"
        "    person: scopatz, appointment: s20, grant: sym,\n"
        "            from 2020-01-01 until 2020-05-15\n"
        "appointments on depleted grants:\n"
        "    person: scopatz, appointment: f19, grant: dmref15,\n"
        "            from 2019-02-07 until 2019-03-31\n"
        "    person: scopatz, appointment: ss20, grant: abc42,\n"
        "            from 2020-07-15 until 2020-08-31\n"
        "    person: scopatz, appointment: ss21, grant: future_grant,\n"
        "            from 2020-09-06 until 2021-08-31\n"
        "underspent grants:\n"
        "    dmref15: end: 2019-05-01\n"
        "      projected underspend: 54.55 months, balance as of 2020-08-31: 0\n"
        "      required ss+gra burn: -3.41\n"
        "    sym: end: 2030-12-31\n"
        "      projected underspend: 8.0 months, balance as of 2020-08-31: 0\n"
        "      required ss+gra burn: 0.06\n"
        "cumulative underspend = 62.55 months, cumulative months to support = 0\n"
        "overspent grants:\n"
        "    end: 2020-12-31, grant: abc42, overspend amount: -1.41 months\n"
        "    end: 2026-08-30, grant: future_grant, overspend amount: -11.97 months\n",
    ),
    (
        [
            "helper",
            "a_proprev",
            "A. Einstein",
            "nsf",
            "2020-04-08",
            "-q",
            "Tess Guebre",
            "--status",
            "downloaded",
            "--title",
            "A flat world theory",
        ],
        "A. Einstein proposal has been added/updated in proposal reviews\n",
    ),
    (
        [
            "helper",
            "a_manurev",
            "Einstein",
            "2020-09-15",
            "Nature",
            "On the Quantum Theory of Radiation",
            "--requester",
            "Niels Bohr",
            "--reviewer",
            "zcliu",
            "--status",
            "submitted",
            "--submitted-date",
            "2019-01-01",
        ],
        "Einstein manuscript has been added/updated in manuscript reviews\n",
    ),
    (
        [
            "helper",
            "a_grppub_readlist",
            "pdf",
            "--title",
            "A list to test the lister",
            "--purpose",
            "Test the lister",
            "--date",
            "2021-04-01",
        ],
        "Making lists for tags:\n['pdf']\npdf has been added/updated in reading_lists\n",
    ),
    (
        [
            "helper",
            "a_proposal",
            "a new proposal",
            "100.0",
            "To destroy numbers",
            "--begin-date",
            "2020-09-15",
            "--end-date",
            "2022-02-14",
            "--duration",
            "16.89",
            "--authors",
            "Kurt Godel",
            "MC Escher",
            "Johann Sebastian Bach",
            "--currency",
            "Bitcoin",
            "--other-agencies",
            "Flatland",
            "--notes",
            "this is a sample added proposal",
            "--date",
            "2020-08-01",
        ],
        "20_anewproposal has been added in proposals\n",
    ),
    # This now tested in the test_helper_python_mock function, below
    # (["helper", "a_expense", "timbuktoo", "travel to timbuktoo", "--amount", "159.18",
    #   "--grants", "mrsec14", "dmref15", "--payee", "ashaaban",
    #   "--where", "bank", "--begin-date", "2020-06-20", "--end-date", "2020-06-25"],
    #  "2006as_timbuktoo has been added in expenses\n"),
    # The following Test Cases A-D test adding presentation-related expenses and map to user stories for Issue #910. All except one are commented out
    # because the current testing architecture (1) limits our ability to validate the addition of more than one entry to a collection, and
    # (2) only spins up one test database, but two would be needed to test a different destination database for expense data. The hope is
    # that, in the future when the test architecture is improved or changed, these commented-out tests can be useful and enable fully testing the added functionality.
    # Test Case A: Expect a new entry in outputs/presentations/presentations.yaml
    # (["helper", "a_presentation", "flat earth", "Mars", "2020-06-26", "2020-06-26",
    #   "--type", "contributed_oral", "--person", "ashaaban", "--grants", "mrsec14",
    #   "--authors", "sbillinge", "ashaaban", "--abstract", "the earth is round as seen from mars",
    #   "--title", "On the roundness of the Earth", "--status", "in-prep",
    #   "--notes", "this is a sample added presentation",
    #   "--presentation-url", "http://drive.google.com/SEV356DV",
    #   "--no-cal", "--no-expense"],
    #  "2006as_mars has been added in presentations\n"),
    # Test Case B: user arguments contradict, raises error
    # (["helper", "a_presentation", "Test Case B", "Test B", "2020-06-26", "2020-06-26",
    #   "--type", "contributed_oral", "--person", "nasker", "--grants", "testing",
    #   "--authors", "sbillinge", "nasker", "--abstract", "testing",
    #   "--title", "Testing Case B", "--status", "in-prep",
    #   "--notes", "This is to test Case B, where user contradicts themselves by passing both --no-expense and --expense_db",
    #   "--presentation-url", "http://drive.google.com/SEV356DV",
    #   "--no-cal", "--no-expense", "--expense-db testB"],
    #  pytest.raises(RuntimeError)),
    # Test Case C.1: user wants an expense added, but did not specify an expense db, and default is public
    # (["helper", "a_presentation", "Test Case C.1", "Test C.1", "2020-06-26", "2020-06-26",
    #   "--type", "contributed_oral", "--person", "nasker", "--grants", "testing",
    #   "--authors", "sbillinge", "nasker", "--abstract", "testing",
    #   "--title", "Testing Case C.1", "--status", "in-prep",
    #   "--notes", "This is to test Case C.1, where user wants an expense added, but did not specify an expense db, and the first db in the regolithrc.json file is public, so the program errors.",
    #   "--presentation-url", "http://drive.google.com/SEV356DV",
    #   "--no-cal"],
    #  pytest.raises(RuntimeError)),
    # Test Case C.2: user wants an expense added, and passed the force option without specifying an expense db, and default is public
    # This is tested in the test_helper_python_mock function, below
    # (["helper", "a_presentation", "Test Case C.2", "Test C.2", "2020-06-26", "2020-06-26",
    #   "--type", "contributed_oral", "--person", "nasker", "--grants", "testing",
    #   "--authors", "sbillinge", "nasker", "--abstract", "testing",
    #   "--title", "Testing Case C.2", "--status", "in-prep",
    #   "--notes", "This is to test Case C.2 where user wants an expense added and passed the force option without specifying an expense db when default is public",
    #   "--presentation-url", "http://drive.google.com/SEV356DV",
    #   "--no-cal", "--force", "--no-repo"], # Expect a new presentation and new expense in db 'test'.
    #    "2006na_testc.2 has been added in presentations\n2006na_testc.2 has been added in expenses in database test\n"),
    # Test Case D: user wants an expense added, and specified an expense db
    # (["helper", "a_presentation", "Test Case D", "Test D", "2020-06-26", "2020-06-26",
    #   "--type", "contributed_oral", "--person", "nasker", "--grants", "testing",
    #   "--authors", "sbillinge", "nasker", "--abstract", "testing",
    #   "--title", "Testing Case D", "--status", "in-prep",
    #   "--notes", "This is to test Case D, where user wants an expense added, and specified an expense-db",
    #   "--presentation-url", "http://drive.google.com/SEV356DV",
    #   "--no-cal", "--expense-db private-test"], # Expect a new presentation and new expense in db 'private-test'
    #    "2006na_testd has been added in presentations\n2006na_testd has been added in expenses in database private-test\n"),
]

db_srcs = [
    # "mongo",
    "fs"
]

# helper_map = [
#     (["helper", "a_grppub_readlist", "all", "all"
#      ],
#      "List of all tags in citations collection:\n['nomonth', 'pdf']\nbuilding lists for all tags in the citation collection\nnomonth has been added/updated in reading_lists\npdf has been added/updated in reading_lists\n"),
#     ]

helper_map_bad = [
    (
        ["helper", "u_milestone", "--milestone_uuid", "sb_fir", "--projectum_id", "sb_firstprojectum"],
        "Detected both a uuid fragment and projectum id.\n"
        "You may enter either a milestone uuid or a projectum id but not both.\n"
        "Enter a milestone uuid to update an existing milestone, or a projectum id to add a new milestone to that projectum.\n",
        RuntimeError,
    ),
    (
        [
            "helper",
            "u_milestone",
            "--due-date",
            "2020-06-01",
            "--notes",
            "do this",
            "do that",
            "--type",
            "mergedpr",
        ],
        "No milestone uuid or projectum id was entered.\n"
        "Enter a milestone uuid to update an existing milestone, or a projectum id to add a new milestone to that projectum.\n",
        RuntimeError,
    ),
    (
        ["helper", "u_milestone", "--projectum_id", "sb_firstprojectum", "--due_date", "2020-06-01"],
        "name, objective, and due date are required for a new milestone",
        RuntimeError,
    ),
    (
        ["helper", "u_milestone", "--milestone_uuid", "milestone_uuid_pl1", "--due_date", "2020-06-01"],
        "Milestone (milestone_uuid_pl1) does not have a type set and this is required.\n"
        "Specify '--type' and rerun the helper to update this milestone.\n",
        ValueError,
    ),
    (
        ["helper", "u_milestone", "--milestone_uuid", "kopl_first", "--type", "bad_type"],
        "The type you have specified is not recognized. \n"
        "Please rerun your command adding '--type' \n"
        "and giving a type from this list:\n"
        f"{MILESTONE_TYPES}\n",
        ValueError,
    ),
    (
        [
            "helper",
            "u_milestone",
            "--projectum_id",
            "sb_firstprojectum",
            "-u",
            "2020-06-01",
            "-n",
            "new milestone",
            "-o",
            "complete ms",
            "--type",
            "bad_type",
        ],
        "The type you have specified is not recognized. \n"
        "Please rerun your command adding '--type' \n"
        "and giving a type from this list:\n"
        f"{MILESTONE_TYPES}\n",
        ValueError,
    ),
    (
        ["helper", "u_milestone", "--milestone_uuid", "milestone_uuid_sb1_", "--type", "bad_type"],
        "The type you have specified is not recognized. \n"
        "Please rerun your command adding '--type' \n"
        "and giving a type from this list:\n"
        f"{MILESTONE_TYPES}\n",
        ValueError,
    ),
    (
        ["helper", "a_todo", "test a_todo", "6", "50", "--assigned-to", "sbillinge", "--milestone_uuid", "bad_id"],
        "No milestone ids were found that match your entry (bad_id).\n"
        "Make sure you have entered the correct milestone uuid or uuid fragment and rerun the helper.",
        RuntimeError,
    ),
    (
        [
            "helper",
            "a_todo",
            "test a_todo",
            "6",
            "50",
            "--assigned-to",
            "sbillinge",
            "--milestone_uuid",
            "one_uuid",
        ],
        "Multiple milestone ids match your entry (one_uuid).\n"
        "Try entering more characters of the uuid and rerunning the helper.",
        RuntimeError,
    ),
    (
        ["helper", "a_todo", "test a_todo", "6", "50", "--assigned-to", "sbillinge", "--milestone_uuid", "_sb1"],
        "Multiple milestone ids match your entry (_sb1).\n"
        "Try entering more characters of the uuid and rerunning the helper.",
        RuntimeError,
    ),
]


@pytest.mark.parametrize("hmb", helper_map_bad)
def test_helpers_bad(hmb, make_db):
    repo = Path(make_db)
    os.chdir(repo)
    with pytest.raises(hmb[2]) as excinfo:
        main(args=hmb[0])
    assert str(excinfo.value) == hmb[1]


@pytest.mark.parametrize("db_src", db_srcs)
@pytest.mark.parametrize("hm", helper_map)
def test_helper_python(hm, make_db, db_src, make_mongodb, capsys, mocker):
    mocker.patch("uuid.uuid4", return_value="test-uuid")
    testfile = Path(__file__)

    if db_src == "fs":
        repo = Path(make_db)
    elif db_src == "mongo":
        if make_mongodb is False:
            pytest.skip("Mongoclient failed to start")
        else:
            repo = Path(make_mongodb)
    os.chdir(repo)

    main(args=hm[0])
    out, err = capsys.readouterr()
    assert hm[1] == out

    expecteddir = testfile.parent / "outputs" / hm[0][1]

    if expecteddir.is_dir():
        if db_src == "fs":
            test_dir = repo / "db"
            assert_outputs(test_dir, expecteddir)
        elif db_src == "mongo":
            from regolith.database import connect
            from regolith.runcontrol import DEFAULT_RC, load_rcfile

            os.chdir(repo)
            rc = copy.copy(DEFAULT_RC)
            rc._update(load_rcfile("regolithrc.json"))
            with connect(rc) as client:
                mongo_database = client[rc.databases[0]["name"]]
                assert_mongo_vs_yaml_outputs(expecteddir, mongo_database)


helper_map_loose = [
    (
        ["helper", "l_abstract"],
        "-------------------------------------------\n"
        "please rerun specifying at least one filter\n"
        "-------------------------------------------\n",
    ),
]


@pytest.mark.parametrize("hm", helper_map_loose)
def test_helper_python_loose(hm, make_db, capsys):
    repo = Path(make_db)
    os.chdir(repo)

    main(args=hm[0])
    out, err = capsys.readouterr()
    assert hm[1] in out


helper_map_requests = [
    (
        [
            "helper",
            "a_expense",
            "timbuktoo",
            "travel to timbuktoo",
            "--amount",
            "159.18",
            "--grants",
            "mrsec14",
            "dmref15",
            "--payee",
            "ashaaban",
            "--where",
            "bank",
            "--begin-date",
            "2020-06-20",
            "--end-date",
            "2020-06-25",
        ],
        "2006as_timbuktoo has been added in expenses\n",
    ),
    (
        [
            "helper",
            "a_presentation",
            "Test Case C.2",
            "Test C.2",
            "2020-06-26",
            "2020-06-26",
            "--type",
            "contributed_oral",
            "--person",
            "nasker",
            "--grants",
            "testing",
            "--authors",
            "sbillinge",
            "nasker",
            "--abstract",
            "testing",
            "--title",
            "Testing Case C.2",
            "--status",
            "in-prep",
            "--notes",
            "This is to test Case C.2 where user wants an expense added and passed the force option without specifying an expense db when default is public",
            "--presentation-url",
            "http://drive.google.com/SEV356DV",
            "--no-cal",
            "--force",
        ],  # Expect a new presentation and new expense in db 'test'
        "2006na_testc.2 has been added in presentations\n2006na_testc.2 has been added in expenses in database test\nrepo 2006na_testc.2 has been created at https://example.com.\nClone this to your local using (HTTPS):\ngit clone https://example.com:talks/2006na_testc.2.git\nor (SSH):\ngit clone git@example.com:talks/2006na_testc.2.git\n",
    ),
]


@pytest.mark.parametrize("db_src", db_srcs)
@pytest.mark.parametrize("hmr", helper_map_requests)
@requests_mock.Mocker(kw="mock")
def test_helper_python_mock(hmr, make_db, db_src, make_mongodb, capsys, **kwargs):
    testfile = Path(__file__)

    kwargs["mock"].post(
        "https://example.com/url/example?namespace_id=35&initialize_with_readme=true&name=2006na_testc.2"
    )

    if db_src == "fs":
        repo = Path(make_db)
    elif db_src == "mongo":
        if make_mongodb is False:
            pytest.skip("Mongoclient failed to start")
        else:
            repo = Path(make_mongodb)
    os.chdir(repo)

    main(args=hmr[0])
    out, err = capsys.readouterr()
    assert hmr[1] == out

    expecteddir = testfile.parent / "outputs" / hmr[0][1]

    if expecteddir.is_dir():
        if db_src == "fs":
            test_dir = repo / "db"
            assert_outputs(test_dir, expecteddir)
        elif db_src == "mongo":
            from regolith.database import connect
            from regolith.runcontrol import DEFAULT_RC, load_rcfile

            os.chdir(repo)
            rc = copy.copy(DEFAULT_RC)
            rc._update(load_rcfile("regolithrc.json"))
            with connect(rc) as client:
                mongo_database = client[rc.databases[0]["name"]]
                assert_mongo_vs_yaml_outputs(expecteddir, mongo_database)


def assert_mongo_vs_yaml_outputs(expecteddir, mongo_database):
    from regolith.dates import convert_doc_iso_to_date
    from regolith.fsclient import load_yaml
    from regolith.mongoclient import load_mongo_col

    os.chdir(expecteddir)
    for root, dirs, files in os.walk("."):
        for file in files:
            fn2 = expecteddir / root / file
            expected_collection_dict = load_yaml(fn2)
            base, ext = os.path.splitext(file)
            mongo_coll_pointer = mongo_database[base]
            edited_collection_dict = load_mongo_col(mongo_coll_pointer)
            for k, v in edited_collection_dict.items():
                edited_collection_dict[k] = convert_doc_iso_to_date(v)
            for k, v in expected_collection_dict.items():
                expected_collection_dict[k] = convert_doc_iso_to_date(v)
            assert edited_collection_dict == expected_collection_dict


def assert_outputs(builddir, expecteddir):
    """

    Parameters
    ----------
    builddir pathlib.Path object
      the directory where the helper has built the output
    expecteddir pathlib.Path object
      the directory where the expected output is found
    """
    os.chdir(builddir)
    for root, dirs, files in os.walk("."):
        for file in files:
            if file in os.listdir(expecteddir / root):
                fn1 = builddir / root / file
                with fn1.open(mode="r") as f:
                    actual = f.read()
                fn2 = expecteddir / root / file
                with fn2.open(mode="r") as f:
                    expected = f.read()

                # Skip because of a date time in
                if file != "rss.xml":
                    # Fixme proper fix for testing hard coded filepaths on
                    # windows
                    if os.name == "nt":
                        if "tmp" not in expected:
                            if "../.." not in expected:
                                assert expected == actual
                    else:
                        assert expected == actual



import calendar
from datetime import date, timedelta

import dateutil.parser as date_parser
import matplotlib.pyplot as plt
import numpy as np

from regolith.dates import get_dates
from regolith.fsclient import _id_key
from regolith.helpers.basehelper import DbHelperBase
from regolith.sorters import position_key
from regolith.tools import all_docs_from_collection, fuzzy_retrieval, get_appointments, merge_collections_superior

# print([k for k,v in chained_db['people'].items()])

MONTH_COST = 3400


def daterange(date1, date2):
    return [date1 + timedelta(n) for n in range(int((date2 - date1).days) + 1)]


def subparser(subpi):
    subpi.add_argument("-g", "--grant", help="grant id for the grant you want to find appointments.")
    subpi.add_argument(
        "-b",
        "--begin-date",
        help="effort reporting period begins on this date, format YYYY-MM-DD. "
        "Attestations run from the beginning to the end of the grant.",
    )
    subpi.add_argument(
        "-e",
        "--end-date",
        help="effort reporting period ends on this date. "
        "Attestations run from the beginning to the end of the grant.",
    )
    subpi.add_argument(
        "--effort-reporting",
        action="store_true",
        help="list loadings by grant per month per person across " "all grants.",
    )
    subpi.add_argument("--verbose", action="store_true", help="increase verbosity.  Show who incurred expenses")
    subpi.add_argument("--no-plot", action="store_true", help="suppress the plotting")
    return subpi


class AttestationsHelper(DbHelperBase):
    """Helper for attestations"""

    # btype must be the same as helper target in helper.py
    btype = "attestation"
    needed_colls = ["people", "grants", "proposals", "expenses"]

    def construct_global_ctx(self):
        """Constructs the global context"""
        super().construct_global_ctx()
        gtx = self.gtx
        rc = self.rc
        # if not rc.database:
        #     rc.database = rc.databases[0]["name"]
        gtx["people"] = sorted(
            all_docs_from_collection(rc.client, "people"),
            key=position_key,
            reverse=True,
        )
        gtx["expenses"] = sorted(all_docs_from_collection(rc.client, "expenses"), key=_id_key)
        gtx["grants"] = sorted(all_docs_from_collection(rc.client, "grants"), key=_id_key)
        gtx["proposals"] = sorted(all_docs_from_collection(rc.client, "proposals"), key=_id_key)
        gtx["all_docs_from_collection"] = all_docs_from_collection
        gtx["float"] = float
        gtx["str"] = str
        gtx["zip"] = zip

    def db_updater(self):
        rc = self.rc
        expenses = self.gtx["expenses"]
        people = self.gtx["people"]
        grants = merge_collections_superior(self.gtx["proposals"], self.gtx["grants"], "proposal_id")
        if not rc.grant and not rc.effort_reporting:
            raise RuntimeError(
                "Grant needed for Attestation.  Please rerun specifying "
                "--grant GRANT_ALIAS or specify --effort-reporting if you "
                "want an effort report."
            )
        if rc.effort_reporting and not rc.begin_date:
            print(rc.begin_date)
            raise RuntimeError(
                "Begin date needed for effort reporting.  Please rerun specifying "
                "--begin-date YYYY-MM-DD (and preferrably --end-date YYYY-MM-DD) "
                "or remove --effort-reporting if you "
                "want an attestation report for a grant"
            )
        if rc.effort_reporting and rc.grant:
            raise print(
                "WARNING: Ignoring the value specified for --grant as "
                "--effort reporting was specified which is over"
                "all grants."
            )
        grant = fuzzy_retrieval(grants, ["name", "_id", "alias"], rc.grant)
        if rc.effort_reporting:
            begin_date = date_parser.parse(rc.begin_date).date()
            if rc.end_date:
                end_date = date_parser.parse(rc.end_date).date()
            else:
                end_date = date.today()
        else:
            print(
                f"Instructions/Notes:\n"
                f"  Quarters are: Q1 July thru Sept, Q2 Oct - Dec, Q3 Jan - Mar, Q4 Apr - Jun\n"
                f"  Grad salaries are about ${MONTH_COST} per month"
            )
            print(f"Collecting Appointments for grant {rc.grant}:")
            begin_date = get_dates(grant).get("begin_date")
            end_date = get_dates(grant).get("end_date")

        plot_date_list = daterange(begin_date, end_date)
        months = [plot_date_list[0]]
        for day in plot_date_list:
            if day.day == 1:
                months.append(day)
        # only if period starts at the beginning of the month the first month will be
        # double counted
        if months[0].day == 1:
            months = months[1:]
        appts = []
        if rc.effort_reporting:
            for person in people:
                if person.get("appointments"):
                    appts = get_appointments(person, appts)
        else:
            for person in people:
                if person.get("appointments"):
                    appts = get_appointments(person, appts, rc.grant)
        appts = [app for app in appts if app[1] < end_date and app[2] >= begin_date]
        appts.sort(key=lambda x: (x[0], x[1]))
        folks = [app[0] for app in appts]
        folks = list(set(folks))
        if not rc.effort_reporting:
            for app in appts:
                print(
                    f"{app[0]}, from {app[1].strftime('%Y-%m-%d')} to "
                    f"{app[2].strftime('%Y-%m-%d')}, loading {app[3]}. Total months: "
                    f"{app[4]}"
                )
        plots, people_loadings = [], []
        loadingc = np.zeros(len(plot_date_list))
        for folk in folks:
            fig, ax = plt.subplots()
            loadinga = np.zeros(len(plot_date_list))
            for app in appts:
                if app[0] == folk:
                    loadingl = []
                    for day in plot_date_list:
                        if app[1] <= day <= app[2]:
                            loadingl.append(app[3])
                        else:
                            loadingl.append(0)
                    loadinga = loadinga + np.array(loadingl)
            loadingc = loadingc + loadinga

            loadingm, accum, days = [], 0, 0
            for day, load in zip(plot_date_list, loadinga):
                accum = accum + load
                days += 1
                if day.day == calendar.monthrange(day.year, day.month)[-1]:
                    loadingm.append(accum * MONTH_COST / days)
                    accum, days = 0, 0
            # capture any partial month at the end of the period
            if days > 0:
                loadingm.append(accum * MONTH_COST / days)

            people_loadings.append((folk, loadinga, loadingm))
            if not rc.no_plot:
                ax.plot_date(plot_date_list, loadinga, ls="-", marker="", label=folk)
                ax.set_xlabel("date")
                ax.set_ylabel(f"loading for student {app[0]}")
                ax.legend(loc="best")
                fig.autofmt_xdate()
                plots.append(fig)

        if not rc.no_plot:
            fig, ax = plt.subplots()
            ax.plot_date(plot_date_list, loadingc, ls="-", marker="")

        if rc.effort_reporting:
            toprow = ","
            for month in months:
                toprow = f"{toprow}, {month.isoformat()}"
            print(toprow)
            for appt in appts:
                row = f"{appt[0]}, {appt[5]}"
                for month in months:
                    if month >= appt[1] and month < appt[2]:
                        row = f"{row}, {appt[3]*100}"
                    else:
                        row = f"{row}, 0.0"
                print(row)
        else:
            print("\n-----------\nLoadings by month\n------------")
            index = 0

            for month in months:
                print(f"{month.isoformat()}:")
                for person in people_loadings:
                    if person[2][index] > 0:
                        print(f"    {person[0]}\tloading: {round(person[2][index], 2)}")
                index += 1
            print("\n----------------\nExpenses\n----------------")
            expenses_on_grant = [expense for expense in expenses if rc.grant in expense.get("grants")]

            if len(expenses_on_grant) > 1:
                expenses_on_grant.sort(key=lambda x: get_dates(x).get("end_date"))
            for expense in expenses_on_grant:
                for reimb in expense.get("reimbursements"):
                    if reimb.get("amount") == 0:
                        amt = 0
                        for exp_item in expense.get("itemized_expenses", []):
                            amt += exp_item.get("unsegregated_expense")
                            amt += exp_item.get("prepaid_expense", 0)
                        reimb["amount"] = amt

            total_spend, month_spend, all_reimb_dates, all_reimb_amts = 0, 0, [], []
            for e in expenses_on_grant:
                reimb_amts = [round(i.get("amount"), 2) for i in e.get("reimbursements", [{}])]
                reimb_dates = [
                    get_dates(i).get("date", get_dates(e).get("end_date")) for i in e.get("reimbursements", [{}])
                ]
                all_reimb_dates.extend(reimb_dates)
                all_reimb_amts.extend(reimb_amts)
                total_spend += sum(reimb_amts)

                if rc.verbose:
                    for reim_date, amt in zip(reimb_dates, reimb_amts):
                        print(
                            f"{reim_date} (reimb date), {get_dates(e).get('end_date')} (expense date): amount: "
                            f"{amt}, "
                        )
                    print(f"  payee: {e.get('payee')} " f"purpose: {e.get('overall_purpose')[:60]}")
            for month in months:
                if month >= begin_date:
                    month_spend = 0
                    for amt, dte in zip(all_reimb_amts, all_reimb_dates):
                        if month.year == dte.year and month.month == dte.month:
                            month_spend += amt
                    print(f"{month}: expenses monthly total = {month_spend:.2f}")

            print(f"Total spend = {round(total_spend, 2)}")

        for plot in plots:
            plt.show()



"""Helper for finding and listing contacts from the contacts.yml database.
Prints name, institution, and email (if applicable) of the contact.
"""

import dateutil
import dateutil.parser as date_parser
from gooey import GooeyParser

from regolith.dates import get_dates, is_current
from regolith.fsclient import _id_key
from regolith.helpers.basehelper import SoutHelperBase
from regolith.tools import all_docs_from_collection, fuzzy_retrieval, get_pi_id, search_collection

TARGET_COLL = "contacts"
HELPER_TARGET = "l_contacts"


def subparser(subpi):
    date_kwargs = {}
    int_kwargs = {}
    if isinstance(subpi, GooeyParser):
        date_kwargs["widget"] = "DateChooser"
        int_kwargs["widget"] = "IntegerField"
    else:
        subpi.add_argument(
            "run", help='run the lister. To see allowed optional arguments, type "regolith helper l_contacts".'
        )
    subpi.add_argument("-v", "--verbose", action="store_true", help="Increases the verbosity of the output.")
    subpi.add_argument(
        "-n", "--name", help="name or name fragment (single argument only) to use to find contacts."
    )
    subpi.add_argument(
        "-i",
        "--inst",
        help="institution or an institution fragment (single argument only) to use to find contacts.",
    )
    subpi.add_argument(
        "-d",
        "--date",
        help="approximate date corresponding to when the contact was entered in the database. "
        "Comes with a default range of 4 months centered around the date; change range using --range argument.",
        **date_kwargs,
    )
    subpi.add_argument(
        "-r",
        "--range",
        help="range (in months) centered around date d specified by --date, i.e. (d +/- r/2).",
        default=4,
        **int_kwargs,
    )
    subpi.add_argument(
        "-o", "--notes", help="fragment (single argument only) to be found in the notes section of a contact."
    )
    subpi.add_argument("-f", "--filter", nargs="+", help="Search this collection by giving key element pairs.")
    subpi.add_argument(
        "-k",
        "--keys",
        nargs="+",
        help="Specify what keys to return values from when running --filter. "
        "If no argument is given the default is just the id.",
    )
    return subpi


class ContactsListerHelper(SoutHelperBase):
    """Helper for finding and listing contacts from the contacts.yml file"""

    # btype must be the same as helper target in helper.py
    btype = HELPER_TARGET
    needed_colls = [f"{TARGET_COLL}", "institutions"]

    def construct_global_ctx(self):
        """Constructs the global context"""
        super().construct_global_ctx()
        gtx = self.gtx
        rc = self.rc
        if "groups" in self.needed_colls:
            rc.pi_id = get_pi_id(rc)
        rc.coll = f"{TARGET_COLL}"
        colls = [
            sorted(all_docs_from_collection(rc.client, collname), key=_id_key) for collname in self.needed_colls
        ]
        for db, coll in zip(self.needed_colls, colls):
            gtx[db] = coll
        gtx["all_docs_from_collection"] = all_docs_from_collection
        gtx["float"] = float
        gtx["str"] = str
        gtx["zip"] = zip

    def sout(self):
        rc = self.rc
        list_search = []
        collection = self.gtx["contacts"]
        if rc.name:
            list_search.extend(["name", rc.name])
        if rc.inst:
            list_search.extend(["institution", rc.inst])
        if rc.notes:
            list_search.extend(["notes", rc.notes])
        if rc.filter:
            list_search.extend(rc.filter)
        filtered_contacts_id = (search_collection(collection, list_search)).strip("    \n")
        filtered_contacts_id = list(filtered_contacts_id.split("    \n"))
        if rc.date:
            date_list = []
            temp_dat = date_parser.parse(rc.date).date()
            temp_dict = {
                "begin_date": (temp_dat - dateutil.relativedelta.relativedelta(months=int(rc.range))).isoformat(),
                "end_date": (temp_dat + dateutil.relativedelta.relativedelta(months=int(rc.range))).isoformat(),
            }
            for contact in collection:
                curr_d = get_dates(contact)["date"]
                if is_current(temp_dict, now=curr_d):
                    date_list.append(contact.get("_id"))
                filtered_contacts_id = [value for value in filtered_contacts_id if value in date_list]
        filtered_contacts = []
        string_contacts = ""
        for contact in collection:
            if contact.get("_id") in filtered_contacts_id:
                filtered_contacts.append(contact)
                institution = contact.get("institution")
                institution_name = fuzzy_retrieval(self.gtx["institutions"], ["name", "_id", "aka"], institution)
                if institution_name:
                    contact["institution"] = institution_name.get("name")
                if rc.verbose:
                    contact_str = f"{contact.get('name')}\n"
                    for k in ["_id", "email", "institution", "department", "notes", "aka"]:
                        if contact.get(k):
                            if isinstance(contact.get(k), list):
                                lst_expanded = "\n        -".join(map(str, contact.get(k)))
                                contact_str += f"    {k}:\n        -{lst_expanded}\n"
                            else:
                                contact_str += f"    {k}: {contact.get(k)}\n"
                    string_contacts += contact_str
                else:
                    string_contacts += (
                        f"{contact.get('name')}  |  {contact.get('_id')}  |"
                        f"  institution: {contact.get('institution')}  |"
                        f"  email: {contact.get('email', 'missing')}\n"
                    )
        print(string_contacts.strip("\n"))
        return



"""Implementation of commands for command line."""

import json
import os
import re
import sys
from copy import copy
from pprint import pprint

from regolith import storage
from regolith.builder import BUILDERS, builder
from regolith.deploy import deploy as dploy
from regolith.emailer import emailer
from regolith.helper import FAST_UPDATER_WHITELIST, HELPERS, UPDATER_HELPERS, helpr
from regolith.runcontrol import RunControl
from regolith.tools import string_types

email = emailer

RE_AND = re.compile(r"\s+and\s+")
RE_SPACE = re.compile(r"\s+")

INGEST_COLL_LU = {".bib": "citations"}


def add_cmd(rc):
    """Adds documents to a collection in a database."""
    docs = [json.loads(doc) if isinstance(doc, string_types) else doc for doc in rc.documents]
    rc.client.insert_many(rc.db, rc.coll, docs)


def _ingest_citations(rc):
    import bibtexparser
    from bibtexparser.bparser import BibTexParser
    from bibtexparser.customization import getnames

    parser = BibTexParser()
    parser.ignore_nonstandard_types = False

    def customizations(record):
        for n in ["author", "editor"]:
            if n in record:
                a = [i for i in record[n].replace("\n", " ").split(", ")]
                b = [i.split(" and ") for i in a]
                c = [item for sublist in b for item in sublist]
                d = [i.strip() for i in c]
                record[n] = getnames(d)

        return record

    parser.customization = customizations
    with open(rc.filename, "r", encoding="utf-8") as f:
        bibs = bibtexparser.load(f, parser=parser)
    for bib in bibs.entries:
        bibid = bib.pop("ID")
        bib["entrytype"] = bib.pop("ENTRYTYPE")
        if "author" in bib:
            bib["author"] = [a.strip() for b in bib["author"] for a in RE_AND.split(b)]
        if "title" in bib:
            bib["title"] = RE_SPACE.sub(" ", bib["title"])
        rc.client.update_one(rc.db, rc.coll, {"_id": bibid}, bib, upsert=True)


def _determine_ingest_coll(rc):
    f = rc.filename
    base, ext = os.path.splitext(f)
    return INGEST_COLL_LU.get(ext, base)


def ingest(rc):
    """Ingests a foreign resource into a database."""
    if rc.coll is None:
        rc.coll = _determine_ingest_coll(rc)
    if rc.coll == "citations":
        _ingest_citations(rc)
    else:
        raise ValueError("don't know how to ingest collection {0!r}".format(rc.coll))


def _run_app(app, rc):
    if hasattr(app, "rc"):
        raise RuntimeError("cannot assign rc to app")
    app.rc = rc
    app.debug = rc.debug
    print("\nDO NOT type Ctrl-C to close the server!!!")
    print("Instead, run the following:")
    print("\n$ curl -d '' http://localhost:5000/shutdown\n")
    app.run(host="localhost")
    del app.rc


def app(rc):
    """Runs flask app"""
    from regolith.app import app

    _run_app(app, rc)


def grade(rc):
    """Runs flask grading app"""
    from regolith.grader import app

    _run_app(app, rc)


def build_db_check(rc):
    """Checks which DBs a builder needs"""
    dbs = set()
    for t in rc.build_targets:
        bldr = BUILDERS[t]
        needed_colls = getattr(bldr, "needed_colls", None)
        # If the requested builder doesn't state DB deps then it requires
        # all dbs!
        if not needed_colls:
            return None
        dbs.update(needed_colls)
    return dbs


def helper_db_check(rc):
    """Checks which DBs a builder needs"""
    # if the helper is an fast_updater, only open the database from rc.database
    rc.fast_updater = False
    for helperkey in UPDATER_HELPERS.keys():
        if helperkey == rc.helper_target and rc.helper_target in FAST_UPDATER_WHITELIST:
            rc.fast_updater = True
    if rc.database is None:
        rc.database = rc.databases[0]["name"]
    if rc.fast_updater:
        rc.databases = [database for database in rc.databases if database.get("name") == rc.database]

    # only open the needed collections
    colls = set()
    bldr = HELPERS[rc.helper_target][0]
    needed_colls = getattr(bldr, "needed_colls", None)
    # If the requested builder doesn't state DB deps then it requires
    # all dbs!
    if not needed_colls:
        return None
    colls.update(needed_colls)
    return colls


def build(rc):
    """Builds all of the build targets"""
    for t in rc.build_targets:
        bldr = builder(t, rc)
        bldr.build()


def helper(rc):
    """Runs the helper targets"""
    hlpr = helpr(rc.helper_target, rc)
    hlpr.hlp()


def deploy(rc):
    """Deploys all of the deployment targets."""
    if not hasattr(rc, "deploy") or len(rc.deploy) == 0:
        raise RuntimeError("run control has no deployment targets!")
    for target in rc.deploy:
        dploy(rc, **target)


def classlist(rc):
    """Sets values for the class list."""
    from regolith.classlist import register

    register(rc)


def json_to_yaml(rc):
    """Converts JSON to YAML"""
    from regolith import fsclient

    for inp in rc.files:
        base, ext = os.path.splitext(inp)
        out = base + ".yaml"
        fsclient.json_to_yaml(inp, out)


def yaml_to_json(rc):
    """Converts YAML to JSON"""
    from regolith import fsclient

    for inp in rc.files:
        base, ext = os.path.splitext(inp)
        out = base + ".json"
        fsclient.yaml_to_json(inp, out)


def fs_to_mongo(rc: RunControl) -> None:
    """Convert database collection from filesystem to mongo db.

    Parameters
    ----------
    rc : RunControl
        The RunControl. The mongo client will be created according to 'mongodbpath' in it. The databases will
        be loaded according to the 'databases' in it.
    """
    from regolith.mongoclient import MongoClient

    client = MongoClient(rc)
    dbs = getattr(rc, "databases")
    for db in dbs:
        client.import_database(db)
    return


def mongo_to_fs(rc: RunControl) -> None:
    """Convert database collection from filesystem to mongo db.

    Parameters
    ----------
    rc : RunControl
        The RunControl. The mongo client will be created according to 'mongodbpath' in it. The databases will
        be loaded according to the 'databases' in it.
    """
    dbs = getattr(rc, "databases")
    for db in dbs:
        rc.client.export_database(db)
    return


def validate(rc):
    """Validate the combined database against the schemas"""
    from regolith.schemas import validate

    print("=" * 10 + "\nVALIDATING\n")
    any_errors = False
    if getattr(rc, "collection"):
        db = {rc.collection: rc.client.chained_db[rc.collection]}
    else:
        db = rc.client.chained_db
    for name, collection in db.items():
        errored_print = False
        for doc_id, doc in collection.items():
            v = validate(name, doc, rc.schemas)
            if v[0] is False:
                if errored_print is False:
                    errored_print = True
                    any_errors = True
                    print(f"Errors found in {name}")
                    print("=" * len(f"Errors found in {name}"))
                print(f"ERROR in {doc_id}:")
                pprint(v[1])
                cap = copy(v[1])
                for vv in v[1]:
                    pprint(doc.get(vv))
                print("-" * 15)
                print("\n")
    if not any_errors:
        print("\nNO ERRORS IN DBS\n" + "=" * 15)
    else:
        # uncomment when debugging scheme errors
        #
        sys.exit(f"Validation failed on some records\n {cap}")
        # sys.exit(f"Validation failed on some records")


DISCONNECTED_COMMANDS = {
    "rc": lambda rc: print(rc._pformat()),
    "deploy": deploy,
    "store": storage.main,
    "json-to-yaml": json_to_yaml,
    "yaml-to-json": yaml_to_json,
}

CONNECTED_COMMANDS = {
    "add": add_cmd,
    "ingest": ingest,
    "app": app,
    "grade": grade,
    "build": build,
    "email": email,
    "classlist": classlist,
    "validate": validate,
    "helper": helper,
    "fs-to-mongo": fs_to_mongo,
    "mongo-to-fs": mongo_to_fs,
}



"""Builder for CVs."""

from copy import deepcopy
from datetime import date

from regolith.builders.basebuilder import LatexBuilderBase
from regolith.fsclient import _id_key
from regolith.sorters import doc_date_key, ene_date_key, position_key
from regolith.stylers import month_fullnames, sentencecase
from regolith.tools import (
    all_docs_from_collection,
    awards_grants_honors,
    dereference_institution,
    filter_employment_for_advisees,
    filter_grants,
    filter_presentations,
    filter_projects,
    filter_publications,
    fuzzy_retrieval,
    make_bibtex_file,
    merge_collections_superior,
    remove_duplicate_docs,
)


class CVBuilder(LatexBuilderBase):
    """Build CV from database entries"""

    btype = "cv"
    needed_colls = ["institutions", "people", "grants", "citations", "projects", "proposals", "presentations"]

    def construct_global_ctx(self):
        """Constructs the global context"""
        super().construct_global_ctx()
        gtx = self.gtx
        rc = self.rc
        gtx["people"] = sorted(
            all_docs_from_collection(rc.client, "people"),
            key=position_key,
            reverse=True,
        )
        gtx["presentations"] = sorted(all_docs_from_collection(rc.client, "presentations"), key=_id_key)
        gtx["institutions"] = sorted(all_docs_from_collection(rc.client, "institutions"), key=_id_key)
        gtx["all_docs_from_collection"] = all_docs_from_collection
        gtx["float"] = float
        gtx["str"] = str
        gtx["zip"] = zip

    def latex(self):
        """Render latex template"""
        rc = self.rc
        gtx = self.gtx
        if rc.people:
            people = [fuzzy_retrieval(gtx["people"], ["aka", "_id", "name"], rc.people[0])]
        else:
            people = gtx["people"]

        for person in people:
            # so we don't modify the dbs when de-referencing
            names = frozenset(person.get("aka", []) + [person["name"]] + [person["_id"]])
            begin_period = date(1650, 1, 1)

            pubs = filter_publications(
                all_docs_from_collection(rc.client, "citations"),
                names,
                reverse=True,
            )
            bibfile = make_bibtex_file(pubs, pid=person["_id"], person_dir=self.bldir)
            emps = person.get("employment", [])
            emps = [em for em in emps if not em.get("not_in_cv", False)]
            for e in emps:
                e["position"] = e.get("position_full", e.get("position").title())
            emps.sort(key=ene_date_key, reverse=True)
            edu = person.get("education", [])
            edu.sort(key=ene_date_key, reverse=True)
            teach = person.get("teaching", [])
            for t in teach:
                t["position"] = t.get("position").title()

            projs = filter_projects(all_docs_from_collection(rc.client, "projects"), names)
            for proj in projs:
                for member in proj.get("team", []):
                    member.get("role", "").replace("pi", "PI")
                    member["role"] = member.get("role", "").title()
            just_grants = list(all_docs_from_collection(rc.client, "grants"))
            just_proposals = list(all_docs_from_collection(rc.client, "proposals"))
            grants = merge_collections_superior(just_proposals, just_grants, "proposal_id")
            presentations = filter_presentations(
                self.gtx["people"],
                self.gtx["presentations"],
                self.gtx["institutions"],
                person.get("_id"),
                statuses=["accepted"],
            )

            for grant in grants:
                for member in grant.get("team"):
                    dereference_institution(member, self.gtx["institutions"])

            pi_grants, pi_amount, _ = filter_grants(grants, names, pi=True)
            coi_grants, coi_amount, coi_sub_amount = filter_grants(grants, names, pi=False)
            for grant in coi_grants:
                format_pi = grant["me"].get("position", "").replace("copi", "Co-PI")
                format_role = format_pi.replace("pi", "PI")
                grant["me"]["position"] = format_role

            aghs = awards_grants_honors(person, "honors")
            service = awards_grants_honors(person, "service", funding=False)
            # TODO: pull this out so we can use it everywhere
            for ee in [emps, edu]:
                for e in ee:
                    dereference_institution(e, self.gtx["institutions"])

            undergrads = filter_employment_for_advisees(
                self.gtx["people"], begin_period, "undergrad", person["_id"]
            )
            for undergrad in undergrads:
                undergrad["role"] = undergrad["role"].title()
            masters = filter_employment_for_advisees(self.gtx["people"], begin_period, "ms", person["_id"])
            for master in masters:
                master["role"] = master["role"].title()
            currents = filter_employment_for_advisees(self.gtx["people"], begin_period, "phd", person["_id"])
            graduateds = filter_employment_for_advisees(self.gtx["people"], begin_period, "phd", person["_id"])
            postdocs = filter_employment_for_advisees(self.gtx["people"], begin_period, "postdoc", person["_id"])
            postdocs = remove_duplicate_docs(postdocs, "name")
            visitors = filter_employment_for_advisees(
                self.gtx["people"], begin_period, "visitor-unsupported", person["_id"]
            )
            visitors = remove_duplicate_docs(visitors, "name")
            for visitor in visitors:
                visitor["role"] = visitor["role"].title()

            iter = deepcopy(graduateds)
            for g in iter:
                if g.get("active"):
                    graduateds.remove(g)
            iter = deepcopy(currents)
            for g in iter:
                if not g.get("active"):
                    currents.remove(g)
            #############
            # hindex
            #############
            if person.get("hindex"):
                hindex = sorted(person.get("hindex"), key=doc_date_key).pop()
            else:
                hindex = {}

            self.render(
                "cv.tex",
                person["_id"] + ".tex",
                p=person,
                title=person.get("name", ""),
                aghs=aghs,
                hindex=hindex,
                service=service,
                undergrads=undergrads,
                masters=masters,
                currentphds=currents,
                graduatedphds=graduateds,
                postdocs=postdocs,
                visitors=visitors,
                pubs=pubs,
                names=names,
                bibfile=bibfile,
                education=edu,
                employment=emps,
                presentations=presentations,
                sentencecase=sentencecase,
                monthstyle=month_fullnames,
                projects=projs,
                pi_grants=pi_grants,
                pi_amount=pi_amount,
                coi_grants=coi_grants,
                coi_amount=coi_amount,
                coi_sub_amount=coi_sub_amount,
            )
            self.pdf(person["_id"])



"""Run Control object for regolith"""

from __future__ import print_function

import io
import json
import os
from collections.abc import Hashable, Iterable, Mapping
from pprint import pformat
from warnings import warn

from regolith.database import connect
from regolith.validators import DEFAULT_VALIDATORS, always_true, noop

FORBIDDEN_NAMES = frozenset(["del", "global"])


def warn_forbidden_name(forname, inname=None, rename=None):
    """Warns the user that a forbidden name has been found."""
    msg = "found forbidden name {0!r}".format(forname)
    if inname is not None:
        msg += " in {0!r}".format(inname)
    if rename is not None:
        msg += ", renaming to {0!r}".format(rename)
    warn(msg, RuntimeWarning)


def ensuredirs(f):
    """For a file path, ensure that its directory path exists."""
    d = os.path.split(f)[0]
    if not os.path.isdir(d):
        os.makedirs(d)


def touch(filename):
    """Opens a file and updates the mtime, like the posix command of the same name."""
    with io.open(filename, "a"):
        os.utime(filename, None)


def exec_file(filename, glb=None, loc=None):
    """A function equivalent to the Python 2.x execfile statement."""
    with io.open(filename, "r") as f:
        src = f.read()
    exec(compile(src, filename, "exec"), glb, loc)


#
# Run Control
#


class NotSpecifiedType(object):
    """A helper class singleton for run control meaning that a 'real' value
    has not been given."""

    def __repr__(self):
        return "NotSpecified"


NotSpecified = NotSpecifiedType()
"""A helper class singleton for run control meaning that a 'real' value
has not been given.
"""


class RunControl(object):
    """A composable configuration class. Unlike argparse.Namespace,
    this keeps the object dictionary (__dict__) separate from the run
    control attributes dictionary (_dict).
    """

    def __init__(self, _updaters=None, _validators=None, **kwargs):
        """Parameters
        -------------
        kwargs : optional
            Items to place into run control.

        """
        self._dict = {}
        self._updaters = _updaters or {}
        self._validators = _validators or {}
        for k, v in kwargs.items():
            setattr(self, k, v)

    def __getattr__(self, key):
        if key in self._dict:
            value = self._dict[key]
        elif key in self.__dict__:
            value = self.__dict__[key]
        elif key in self.__class__.__dict__:
            value = self.__class__.__dict__[key]
        else:
            msg = "RunControl object has no attribute {0!r}.".format(key)
            raise AttributeError(msg)
        if isinstance(value, property):
            value = value.fget(self)
        return value

    def __setattr__(self, key, value):
        if key.startswith("_"):
            self.__dict__[key] = value
        else:
            if value is NotSpecified and key in self:
                return
            value = self._validate(key, value)
            self._dict[key] = value

    def __delattr__(self, key):
        if key in self._dict:
            del self._dict[key]
        elif key in self.__dict__:
            del self.__dict__[key]
        elif key in self.__class__.__dict__:
            del self.__class__.__dict__[key]
        else:
            msg = "RunControl object has no attribute {0!r}.".format(key)
            raise AttributeError(msg)

    def __iter__(self):
        return iter(self._dict)

    def __repr__(self):
        keys = sorted(self._dict.keys())
        s = ", ".join(["{0!s}={1!r}".format(k, self._dict[k]) for k in keys])
        return "{0}({1})".format(self.__class__.__name__, s)

    def _get(self, key, default=None):
        try:
            val = getattr(self, key)
        except (KeyError, AttributeError):
            val = default
        return val

    def _pformat(self):
        keys = sorted(self._dict.keys())
        s = ",\n ".join(map(lambda k: "{0!s}={1}".format(k, pformat(self._dict[k], indent=2)), keys))
        return "{0}({1})".format(self.__class__.__name__, s)

    def __contains__(self, key):
        return key in self._dict or key in self.__dict__ or key in self.__class__.__dict__

    def __eq__(self, other):
        if hasattr(other, "_dict"):
            return self._dict == other._dict
        elif isinstance(other, Mapping):
            return self._dict == other
        else:
            return NotImplemented

    def __ne__(self, other):
        if hasattr(other, "_dict"):
            return self._dict != other._dict
        elif isinstance(other, Mapping):
            return self._dict != other
        else:
            return NotImplemented

    def __copy__(self):
        return type(self)(_updaters=self._updaters, _validators=self._validators, **self._dict)

    def _update(self, other):
        """Updates the rc with values from another mapping.  If this rc has
        if a key is in self, other, and self._updaters, then the updaters
        value is called to perform the update.  This function should return
        a copy to be safe and not update in-place.
        """
        if hasattr(other, "_dict"):
            other = other._dict
        elif not hasattr(other, "items"):
            other = dict(other)
        for k, v in other.items():
            if v is NotSpecified:
                pass
            elif k in self._updaters and k in self:
                v = self._updaters[k](getattr(self, k), v)
            setattr(self, k, v)

    def _validate(self, key, value):
        """Validates - and possibly converts - a value based on its key and the current
        validators.
        """
        validators = self._validators
        if key in validators:
            validator, convertor = validators[key]
        else:
            for vld in validators:
                if isinstance(vld, str):
                    continue
                m = vld.match(key)
                if m is not None:
                    validator, convertor = validators[vld]
            else:
                validator, convertor = always_true, noop
        return value if validator(value) else convertor(value)


def flatten(iterable):
    """Generator which returns flattened version of nested sequences."""
    for el in iterable:
        if isinstance(el, str):
            yield el
        elif isinstance(el, Iterable):
            for subel in flatten(el):
                yield subel
        else:
            yield el


#
# Memoization
#


def ishashable(x):
    """Tests if a value is hashable."""
    if isinstance(x, Hashable):
        if isinstance(x, str):
            return True
        elif isinstance(x, Iterable):
            return all(map(ishashable, x))
        else:
            return True
    else:
        return False


DEFAULT_RC = RunControl(
    _validators=DEFAULT_VALIDATORS,
    builddir="_build",
    mongodbpath=property(lambda self: os.path.join(self.builddir, "_dbpath")),
    user_config=os.path.expanduser("~/.config/regolith/user.json"),
    force=False,
    database=None,
)


def load_json_rcfile(fname):
    """Loads a JSON run control file."""
    with open(fname, "r", encoding="utf-8") as f:
        rc = json.load(f)
    return rc


def load_rcfile(fname):
    """Loads a run control file."""
    base, ext = os.path.splitext(fname)
    if ext == ".json":
        rc = load_json_rcfile(fname)
    else:
        raise RuntimeError("could not detemine run control file type from extension.")
    return rc


def filter_databases(rc):
    """Filters the databases list down to only the ones we need, in place."""
    dbs = rc.databases
    public_only = rc._get("public_only", False)
    if public_only:
        dbs = [db for db in dbs if db["public"]]
    dbname = rc._get("db")
    if dbname is not None:
        dbs = [db for db in dbs if db["name"] == dbname]
    elif len(dbs) == 1:
        rc.db = dbs[0]["name"]
    rc.databases = dbs


def connect_db(rc, colls=None):
    """
    Load up the db's

    Parameters
    ----------
    rc:
        The runcontrol instance
    colls
        The list of collections that should be loaded

    Returns
    -------
    chained_db:
      The chained databases in the form of a document
    dbs:
       The databases in the form of a runcontrol client
    """
    with connect(rc, dbs=colls) as rc.client:
        dbs = rc.client.dbs
        chained_db = rc.client.chained_db
    return chained_db, dbs



"""Helper for listing group members."""

from regolith.dates import get_dates, is_current
from regolith.fsclient import _id_key
from regolith.helpers.basehelper import SoutHelperBase
from regolith.sorters import position_key
from regolith.tools import (
    all_docs_from_collection,
    collection_str,
    fuzzy_retrieval,
    get_pi_id,
    key_value_pair_filter,
)

TARGET_COLL = "people"
HELPER_TARGET = "l_members"
ALLOWED_STATI = ["proposed", "started", "finished", "back_burner", "paused", "cancelled"]


def subparser(subpi):
    subpi.add_argument("-c", "--current", action="store_true", help="get only current group members ")
    subpi.add_argument("-p", "--prior", action="store_true", help="get only former group members ")
    subpi.add_argument("-v", "--verbose", action="store_true", help="increase verbosity of output")
    subpi.add_argument("-f", "--filter", nargs="+", help="Search this collection by giving key element pairs")
    subpi.add_argument(
        "-k",
        "--keys",
        nargs="+",
        help="Specify what keys to return values from when running "
        "--filter. If no argument is given the default is just the id.",
    )

    return subpi


class MembersListerHelper(SoutHelperBase):
    """Helper for listing group members."""

    # btype must be the same as helper target in helper.py
    btype = HELPER_TARGET
    needed_colls = [f"{TARGET_COLL}", "institutions", "groups"]

    def construct_global_ctx(self):
        """Constructs the global context"""
        super().construct_global_ctx()
        gtx = self.gtx
        rc = self.rc
        if "groups" in self.needed_colls:
            rc.pi_id = get_pi_id(rc)
        rc.coll = f"{TARGET_COLL}"
        colls = [
            sorted(all_docs_from_collection(rc.client, collname), key=_id_key) for collname in self.needed_colls
        ]
        for db, coll in zip(self.needed_colls, colls):
            gtx[db] = coll
        gtx["all_docs_from_collection"] = all_docs_from_collection
        gtx["float"] = float
        gtx["str"] = str
        gtx["zip"] = zip

    def sout(self):
        gtx = self.gtx
        rc = self.rc
        if rc.filter:
            collection = key_value_pair_filter(self.gtx["people"], rc.filter)
        else:
            collection = self.gtx["people"]
        people = []
        group = fuzzy_retrieval(gtx["groups"], ["_id", "aka", "name"], rc.groupname)
        group_id = group.get("_id")

        if rc.filter:
            if not rc.verbose:
                results = collection_str(collection, rc.keys)
                print(results, end="")
                return
            else:
                for person in collection:
                    print(
                        "{}, {} | group_id: {}".format(
                            person.get("name"), person.get("position"), person.get("_id")
                        )
                    )
                    print("    orcid: {} | github_id: {}".format(person.get("orcid_id"), person.get("github_id")))
                pass
            # code to print verbosely on filtering
        if not rc.filter:
            for person in gtx["people"]:
                if rc.current:
                    if not person.get("active"):
                        continue
                    people.append(person)
                elif rc.prior:
                    if person.get("active"):
                        continue
                    people.append(person)
                else:
                    people.append(person)

        cleaned_people = []
        for person in people:
            not_current_positions = [emp for emp in person.get("employment") if not is_current(emp)]
            not_current_positions.sort(key=lambda x: get_dates(x)["end_date"])
            current_positions = [emp for emp in person.get("employment") if is_current(emp)]
            current_positions.sort(key=lambda x: get_dates(x)["begin_date"])
            positions = not_current_positions + current_positions
            position_keys = [
                position_key(position) for position in positions if position.get("group", "") == group_id
            ]
            if position_keys:
                person["position_key"] = max(position_keys)[0]
                cleaned_people.append(person)
            else:
                print(f"Person {person['name']} has no positions in group {group_id}")
        cleaned_people.sort(key=lambda k: k["position_key"], reverse=True)
        position_names = {
            1: "Undergrads",
            2.5: "Masters Students",
            2: "Visiting Students",
            3: "Graduate Students",
            4: "Post Docs",
            5: "Visitors",
            8: "Assistant Scientists",
            9: "Associate Scientists",
            10: "Scientists",
            11: "PI",
        }
        accounting = 12
        for person in cleaned_people:
            if person.get("position_key") < accounting:
                accounting = person.get("position_key")
                print(f"    -- {position_names.get(accounting, position_names.get(5))} --")
            if rc.verbose:
                print("{}, {}".format(person.get("name"), person.get("position")))
                print("    email: {} | group_id: {}".format(person.get("email"), person.get("_id")))
                print("    github_id: {} | orcid: {}".format(person.get("github_id"), person.get("orcid_id")))
                for position in positions:
                    if is_current(position):
                        inst = fuzzy_retrieval(
                            gtx["institutions"], ["aka", "name", "_id"], position.get("organization")
                        )
                        if inst:
                            instname = inst.get("name")
                        else:
                            print(f"WARNING: {position.get('organization')} not in institutions collection")
                        print("    current organization: {}".format(instname))
                        print(
                            "    current position: {}".format(
                                position.get("full_position", position.get("position").title())
                            )
                        )
                    if not person.get("active"):
                        if position.get("group") == "bg":
                            print("    billinge group position: {}".format(position.get("position")))
            else:
                print("{}".format(person.get("name")))
        return



"""Builder for Figure including documents."""

import os

from jinja2 import Environment, FileSystemLoader

from regolith.broker import Broker
from regolith.builders.basebuilder import LatexBuilderBase
from regolith.tools import fuzzy_retrieval


class FigureBuilder(LatexBuilderBase):
    """Build resume from database entries"""

    btype = "figure"

    def __init__(self, rc):
        self.env = Environment(loader=FileSystemLoader(["."]))
        self.db = Broker(rc)
        super().__init__(rc)

    def construct_global_ctx(self):
        super().construct_global_ctx()
        gtx = self.gtx
        gtx["db"] = self.db.md
        gtx["get_file_path"] = self.db.get_file_path
        gtx["fuzzy_retrieval"] = fuzzy_retrieval

    def latex(self):
        """Render latex template"""
        for f in [ff for ff in os.listdir(".") if ff.endswith(".tex")]:
            fn, ext = os.path.splitext(f)
            self.render(f, fn + "_rend" + ext)
            self.pdf(fn + "_rend")



"""Builder for the planning of beamtimes.

The plan contains a summary of the information for the experiments
in during a beamtime and details about how to carry out the experiments.

"""

from datetime import datetime

import pandas as pd

from regolith.builders.basebuilder import LatexBuilderBase
from regolith.tools import all_docs_from_collection, group, id_key


class BeamPlanBuilder(LatexBuilderBase):
    """
    Build a file of experiment plans for the beamtime from database entries.

    The report is in the '.tex' file. The template of the file is in the
    'templates/beamplan.tex'. The data will be grouped
    according to beamtime. Each beamtime will generate a file of the plans.
    If 'beamtime' in 'rc' are not None, only plans for those beamtime will be generated.

    Methods
    -------
    construct_global_ctx()
        Constructs the global context.
    latex()
        Render latex template.
    """

    btype = "beamplan"
    needed_colls = ["beamplan", "beamtime"]

    def construct_global_ctx(self):
        """Constructs the global context."""
        super().construct_global_ctx()
        gtx = self.gtx
        rc = self.rc
        gtx["beamplan"] = all_docs_from_collection(rc.client, "beamplan")
        gtx["beamtime"] = all_docs_from_collection(rc.client, "beamtime")
        gtx["all_docs_from_collection"] = all_docs_from_collection

    @staticmethod
    def _to_readable(date):
        """Convert the string date to a human readable form."""
        if date is None:
            return "missing"
        date_obj = datetime.strptime(date, "%Y-%m-%d")
        readable_date = date_obj.strftime("%b %d, %Y")
        return readable_date

    @staticmethod
    def _search(db, key):
        """Search doc in the database."""
        for doc in db:
            if id_key(doc) == key:
                return doc
        return None

    def _gather_info(self, bt, docs):
        """
        Query information from the list of documents. Return a table as the summary of the plans and a list of
        experiment plans.

        Parameters
        ----------
        bt : str
            The name of the beamtime. It should be a key in the beamtime.yml.
        docs : list
            A list of documents of the experiment plans in the beamplan database.

        Returns
        -------
        info : dict
            The information obtained from the database and formatted. It contains the key value pairs:
            table - The latex string of table (used in tex file).
            table_str - The string of the table (used in txt file).
            tasks - The combination of todos in the plans.
            plans - The list of experiment plans. Each item in a plan is a list of strings.
            begin_date - The beginning date of the beamtime.
            end_date - The end date of the beamtime.
            caption - caption of the table.
        """
        # get begin_date and end_date
        beamtime = self.gtx["beamtime"]
        bt_doc = self._search(beamtime, bt)
        if bt_doc:
            begin_date = self._to_readable(bt_doc.get("begin_date"))
            end_date = self._to_readable(bt_doc.get("end_date"))
            begin_time = bt_doc.get("begin_time", "missing")
            end_time = bt_doc.get("end_time", "missing")
        else:
            begin_time = end_time = begin_date = end_date = "no doc"
        # get data from beamplan
        rows, plans, tasks = [], [], []
        docs = sorted(docs, key=lambda d: d.get("devices"))
        for n, doc in enumerate(docs):
            # gather information of the table
            serial_id = str(n + 1)
            row = {
                "serial id": serial_id,
                "project leader": doc.get("project_lead", ""),
                "number of samples": str(len(doc.get("samples", []))),
                "measurement": doc.get("measurement", ""),
                "devices": ", ".join(doc.get("devices", [])),
                "estimated time (h)": "{:.1f}".format(doc.get("time", 0) / 60),
            }
            rows.append(row)
            # gather information of the plan.
            plan = {
                "serial_id": serial_id,
                "samples": doc.get("samples", []),
                "objective": doc.get("objective", ""),
                "prep_plan": doc.get("prep_plan", []),
                "ship_plan": doc.get("ship_plan", []),
                "exp_plan": doc.get("exp_plan", []),
                "scanplan": doc.get("scanplan", []),
            }
            plans.append(plan)
            # gather info of the task
            todo_list = ["(Exp. {}) {}".format(serial_id, todo) for todo in doc.get("todo", [])]
            tasks += todo_list
        # make a pandas dataframe and calculate time and samples
        df = pd.DataFrame(rows)
        total_time = "{:.1f}".format(df["estimated time (h)"].astype(float).sum())
        total_sample_num = "{:d}".format(df["number of samples"].astype(int).sum())
        # convert to the string form
        table_latex = df.to_latex(escape=True, index=False)
        table_str = df.to_string()
        # make a dict
        info = {
            "bt": bt,  # str
            "plans": plans,  # List[dict]
            "table": table_latex,  # str
            "table_for_txt": table_str,  # str
            "tasks": tasks,  # List[str]
            "begin_date": begin_date,  # str
            "end_date": end_date,  # str
            "begin_time": begin_time,  # str
            "end_time": end_time,  # str
            "total_time": total_time,  # str
            "total_sample_num": total_sample_num,  # str
        }
        return info

    def latex(self):
        """Render latex template."""
        gtx = self.gtx
        db = gtx["beamplan"]
        grouped = group(db, "beamtime")
        for bt, plans in grouped.items():
            info = self._gather_info(bt, plans)
            self.render("beamplan.tex", "{}.tex".format(bt), **info)
            self.render("beamplan.txt", "{}.txt".format(bt), **info)
            self.pdf(bt)
        return



"""The main CLI for regolith"""

from __future__ import print_function

import copy
import os
from argparse import ArgumentParser, Namespace, RawTextHelpFormatter

from regolith import __version__, commands
from regolith.builder import BUILDERS
from regolith.commands import CONNECTED_COMMANDS, DISCONNECTED_COMMANDS, INGEST_COLL_LU
from regolith.database import connect
from regolith.helper import HELPERS
from regolith.runcontrol import DEFAULT_RC, filter_databases, load_rcfile
from regolith.schemas import SCHEMAS
from regolith.tools import update_schemas

NEED_RC = set(CONNECTED_COMMANDS.keys())
NEED_RC |= {"rc", "deploy", "store"}


def create_parser():
    p = ArgumentParser()
    subp = p.add_subparsers(title="cmd", dest="cmd")

    p.add_argument("--version", action="store_true")

    # helper subparser
    subp.add_parser(
        "helper",
        help="runs an available helper target",
        formatter_class=RawTextHelpFormatter,
    )

    # rc subparser
    subp.add_parser("rc", help="prints run control")

    # add subparser
    addp = subp.add_parser("add", help="adds a record to a database and collection")
    addp.add_argument("db", help="database name")
    addp.add_argument("coll", help="collection name")
    addp.add_argument("documents", nargs="+", help="documents, in JSON / mongodb format")

    # ingest subparser
    ingp = subp.add_parser(
        "ingest",
        help="ingest many records from a foreign " "resource into a database",
    )
    ingp.add_argument("db", help="database name")
    ingp.add_argument(
        "filename",
        help="file to ingest. Currently valid formats are: \n{}" "".format([k for k in INGEST_COLL_LU]),
    )
    ingp.add_argument(
        "--coll",
        dest="coll",
        default=None,
        help="collection name, if this is not given it is infered from the " "file type or file name.",
    )

    # store subparser
    strp = subp.add_parser("store", help="stores a file into the appropriate " "storage location.")
    strp.add_argument("storename", help="storage name")
    strp.add_argument(
        "documents",
        nargs="+",
        help="paths to documents, i.e. " "PDFs, images, etc.",
    )
    strp.add_argument(
        "-f",
        "--force",
        dest="force",
        default=False,
        action="store_true",
        help="forces copy of file if one of the same name " "already exists",
    )

    # app subparser
    appp = subp.add_parser(
        "app",
        help="starts up a flask app for inspecting and " "modifying regolith data.",
    )
    appp.add_argument(
        "--debug",
        dest="debug",
        action="store_true",
        default=False,
        help="starts server in debug mode",
    )

    # grade subparser
    grdp = subp.add_parser(
        "grade",
        help="starts up a flask app for adding " "grades to the database.",
    )
    grdp.add_argument(
        "--debug",
        dest="debug",
        action="store_true",
        default=False,
        help="starts server in debug mode",
    )

    # builder subparser
    bldp = subp.add_parser(
        "build",
        help="builds various available targets",
        formatter_class=RawTextHelpFormatter,
    )
    bldp.add_argument(
        "build_targets",
        nargs="+",
        help="targets to build. Currently valid targets are: \n{}".format([k for k in BUILDERS]),
    )
    bldp.add_argument(
        "--no-pdf",
        dest="pdf",
        help="don't produce PDFs during the build " "(for builds which produce PDFs)",
        action="store_false",
        default=True,
    )
    bldp.add_argument(
        "--from",
        dest="from_date",
        help="date in form YYYY-MM-DD.  Items will only be built"
        " if their date or end_date is equal or after this date",
        default=None,
    )
    bldp.add_argument(
        "--to",
        dest="to_date",
        help="date in form YYYY-MM-DD.  Items will only be built"
        " if their date or begin_date is equal or before this date",
        default=None,
    )
    bldp.add_argument(
        "--grants",
        nargs="+",
        dest="grants",
        help="specify a grant or a space-separated list of grants so items are "
        "built only if associated with this(these) grant(s)",
        default=None,
    )
    bldp.add_argument(
        "--people",
        nargs="+",
        dest="people",
        help="specify a person or a space-separated list of people such that "
        "the build will be for only those people",
        default=None,
    )
    bldp.add_argument(
        "--kwargs",
        nargs="+",
        dest="kwargs",
        help="pass a specific command to build a specific task " "if it exists",
        default=None,
    )

    # deploy subparser
    subp.add_parser("deploy", help="deploys what was built by regolith")

    # email subparser
    emlp = subp.add_parser("email", help="automates emailing")
    emlp.add_argument("email_target", help='targets to email, eg "test" or ' '"grades".')
    emlp.add_argument("--to", default=None, dest="to", help="receiver of email")
    emlp.add_argument("--subject", dest="subject", help="email subject line", default="")
    emlp.add_argument(
        "--body",
        dest="body",
        help="email body, as restructured text",
        default="",
    )
    emlp.add_argument(
        "--attach",
        nargs="+",
        dest="attachments",
        default=(),
        help="attachments to send along as well.",
    )
    emlp.add_argument(
        "-c",
        "--course-id",
        dest="course_ids",
        default=(),
        nargs="+",
        help="course identifier that should be emailed.",
    )
    emlp.add_argument("--db", help="database name", dest="db", default=None)

    # classlist subparser
    clp = subp.add_parser("classlist", help="updates classlist information from file")
    clp.add_argument("op", help='operatation to perform, such as "add" or "replace".')
    clp.add_argument("filename", help="file to read class information from.")
    clp.add_argument("course_id", help="course identifier whose registry should be updated")
    clp.add_argument(
        "-f",
        "--format",
        dest="format",
        default=None,
        help="file / school format to read information from. Current values are "
        '"json" and "usc". Determined from extension if not available.',
    )
    clp.add_argument(
        "-d",
        "--dry-run",
        dest="dry_run",
        action="store_true",
        default=False,
        help="only does a dry run and reports results",
    )
    clp.add_argument("--db", help="database name", dest="db", default=None)

    # JSON-to-YAML subparser
    jty = subp.add_parser("json-to-yaml", help="Converts files from JSON to YAML")
    jty.add_argument("files", nargs="+", help="file names to convert")

    # YAML-to-JSON subparser
    ytj = subp.add_parser("yaml-to-json", help="Converts files from YAML to JSON")
    ytj.add_argument("files", nargs="+", help="file names to convert")

    # mongo-to-fs subparser
    mtf = subp.add_parser(
        "mongo-to-fs",
        help="Backup database from mongodb to filesystem as json. "
        "The database will be imported to the destination specified by the 'database':'dst_url' key. "
        "For this to work, ensure that the database is included in the "
        "dst_url, and that local is set to true.",
    )

    mtf.add_argument(
        "--host",
        help="Specifies a resolvable hostname for the mongod to which to connect. By "
        "default, the mongoexport attempts to connect to a MongoDB instance running "
        "on the localhost on port number 27017.",
        dest="host",
        default=None,
    )

    # fs-to-mongo subparser
    ftm = subp.add_parser(
        "fs-to-mongo",
        help="Import database from filesystem to mongodb. By default, the database will be import to the local "
        "mongodb. The database can also be imported to the destination specified by the 'database':'dst_url' key."
        " For this to work, ensure that the database is included in the dst_url, and that local is set to true.",
    )

    ftm.add_argument(
        "--host",
        help="Specifies a resolvable hostname for the mongod to which to connect. By "
        "default, the mongoimport attempts to connect to a MongoDB instance running "
        "on the localhost on port number 27017.",
        dest="host",
        default=None,
    )

    # Validator
    val = subp.add_parser("validate", help="Validates db")
    val.add_argument(
        "--collection",
        dest="collection",
        default=None,
        help="If provided only validate that collection",
    )
    return p


def main(args=None):
    rc = copy.copy(DEFAULT_RC)
    parser = create_parser()
    args0 = Namespace()
    args1, rest = parser.parse_known_args(args, namespace=args0)
    if args1.version:
        print(__version__)
        return rc
    if args1.cmd == "helper":
        p = ArgumentParser(prog="regolith helper")
        p.add_argument(
            "helper_target",
            help="helper target to run. Currently valid targets are: \n{}".format([k for k in HELPERS]),
        )
        if len(rest) == 0:
            p.print_help()
        args2, rest2 = p.parse_known_args(rest, namespace=args0)
        # it is not apparent from this but the following line calls the subparser in
        #   in the helper module to get the rest of the args.
        HELPERS[args2.helper_target][1](p)
        if len(rest2) == 0:
            p.print_help()
        args3, rest3 = p.parse_known_args(rest, namespace=args0)
        ns = args3
    else:
        ns = args1
    if ns.cmd in NEED_RC:
        if os.path.exists(rc.user_config):
            rc._update(load_rcfile(rc.user_config))
        rc._update(load_rcfile("regolithrc.json"))
    rc._update(ns.__dict__)
    if "schemas" in rc._dict:
        user_schema = copy.deepcopy(rc.schemas)
        default_schema = copy.deepcopy(SCHEMAS)
        rc.schemas = update_schemas(default_schema, user_schema)
    else:
        rc.schemas = SCHEMAS
    if ns.cmd in NEED_RC:
        filter_databases(rc)
    if rc.cmd in DISCONNECTED_COMMANDS:
        DISCONNECTED_COMMANDS[rc.cmd](rc)
    else:
        dbs = None
        if rc.cmd == "build":
            dbs = commands.build_db_check(rc)
        elif rc.cmd == "helper":
            dbs = commands.helper_db_check(rc)
        with connect(rc, dbs=dbs) as rc.client:
            CONNECTED_COMMANDS[rc.cmd](rc)
    return rc


if __name__ == "__main__":
    main()



import os
import sys
from io import StringIO
from subprocess import CalledProcessError

import pytest
from xonsh.lib import subprocess

from regolith.main import main


def test_validate_python(make_db):
    repo = make_db
    os.chdir(repo)
    backup = sys.stdout
    sys.stdout = StringIO()
    main(["validate"])
    out = sys.stdout.getvalue()
    sys.stdout.close()
    sys.stdout = backup
    assert "NO ERRORS IN DBS" in out


def test_validate_python_single_col(make_db):
    """
    to see output from a failed test, comment out the code that rediriects stdout
    to out and replace the assert with 'assert false'. Change it back afterwards
    """
    repo = make_db
    os.chdir(repo)
    # to see what is failing, comment out the rows that capture and restore the
    # sys.stdout.
    backup = sys.stdout
    sys.stdout = StringIO()
    main(["validate", "--collection", "projecta"])
    out = sys.stdout.getvalue()
    sys.stdout.close()
    sys.stdout = backup
    assert "NO ERRORS IN DBS" in out
    # assert False


def test_validate_bad_python(make_bad_db):
    repo = make_bad_db
    os.chdir(repo)
    backup = sys.stdout
    sys.stdout = StringIO()
    with pytest.raises(SystemExit):
        main(["validate"])
    out = sys.stdout.getvalue()
    sys.stdout.close()
    sys.stdout = backup
    assert "Errors found in " in out
    assert "NO ERRORS IN DBS" not in out


@pytest.mark.skipif(sys.platform == "win32", reason="does not run on windows")
def test_validate(make_db):
    repo = make_db
    os.chdir(repo)
    out = subprocess.check_output(["regolith", "validate"])
    if isinstance(out, bytes):
        out = out.decode("utf-8")
    assert "NO ERRORS IN DBS" in out


@pytest.mark.skipif(sys.platform == "win32", reason="does not run on windows")
def test_validate_bad(make_bad_db):
    repo = make_bad_db
    os.chdir(repo)
    try:
        subprocess.check_output(["regolith", "validate"])
    except CalledProcessError as e:
        assert "Errors found in " in e.output
        assert "NO ERRORS IN DBS" not in e.output



"""Builder for Resumes."""

from regolith.builders.basebuilder import LatexBuilderBase
from regolith.sorters import ene_date_key, position_key
from regolith.tools import (
    all_docs_from_collection,
    awards_grants_honors,
    filter_grants,
    filter_projects,
    filter_publications,
    fuzzy_retrieval,
    latex_safe,
    make_bibtex_file,
    merge_collections_superior,
    month_and_year,
)


class ResumeBuilder(LatexBuilderBase):
    """Build resume from database entries"""

    btype = "resume"
    needed_colls = ["institutions", "people", "grants", "citations", "projects", "proposals"]

    def construct_global_ctx(self):
        """Constructs the global context"""
        super().construct_global_ctx()
        gtx = self.gtx
        rc = self.rc
        gtx["month_and_year"] = month_and_year
        gtx["latex_safe"] = latex_safe
        gtx["people"] = sorted(
            all_docs_from_collection(rc.client, "people"),
            key=position_key,
            reverse=True,
        )
        gtx["all_docs_from_collection"] = all_docs_from_collection

    def latex(self):
        """Render latex template"""
        rc = self.rc
        if rc.people:
            people = [fuzzy_retrieval(self.gtx["people"], ["aka", "_id", "name"], rc.people[0])]
        else:
            people = self.gtx["people"]

        for p in people:
            names = frozenset(p.get("aka", []) + [p["name"]])
            pubs = filter_publications(
                all_docs_from_collection(rc.client, "citations"),
                names,
                reverse=True,
            )
            bibfile = make_bibtex_file(pubs, pid=p["_id"], person_dir=self.bldir)
            emp = p.get("employment", [])
            emp = [em for em in emp if not em.get("not_in_cv", False)]
            for e in emp:
                e["position"] = e.get("position_full", e.get("position").title())
            emp.sort(key=ene_date_key, reverse=True)
            edu = p.get("education", [])
            edu.sort(key=ene_date_key, reverse=True)
            projs = filter_projects(all_docs_from_collection(rc.client, "projects"), names)
            grants = list(all_docs_from_collection(rc.client, "grants"))
            proposals = list(all_docs_from_collection(rc.client, "proposals"))
            grants = merge_collections_superior(proposals, grants, "proposal_id")
            pi_grants, pi_amount, _ = filter_grants(grants, names, pi=True)
            coi_grants, coi_amount, coi_sub_amount = filter_grants(grants, names, pi=False)
            aghs = awards_grants_honors(p, "honors")
            self.render(
                "resume.tex",
                p["_id"] + ".tex",
                p=p,
                title=p.get("name", ""),
                aghs=aghs,
                pubs=pubs,
                names=names,
                bibfile=bibfile,
                education=edu,
                employment=emp,
                projects=projs,
                pi_grants=pi_grants,
                pi_amount=pi_amount,
                coi_grants=coi_grants,
                coi_amount=coi_amount,
                coi_sub_amount=coi_sub_amount,
            )
            self.pdf(p["_id"])
